import pytest
from unittest.mock import Mock, patch
from dagster_aws.s3 import S3Resource
from geojson import Point
from pydantic import BaseModel
from uuid import UUID
from horizon_data_core.helpers import (
    convert_uri_to_url,
    create_predicate,
    dataframe_to_list,
    list_to_dataframe,
    get_center_of_points,
    name_to_uuid,
    uuid_from_s3_object,
)
from pandas import DataFrame
from s3path import S3Path


def test_get_center_of_points() -> None:
    point = {"coordinates": [1.0, 1.0], "type": "Point"}
    assert get_center_of_points([point]) == Point([1.0, 1.0])


mock_region = "us-gov-east-1"
mock_bucket = "bucket"
mock_s3_resource = S3Resource(endpoint_url=f"https://{mock_bucket}.s3.{mock_region}.amazonaws.com")
mock_local_s3_resource = S3Resource(endpoint_url="http://localhost:9000")


def test_convert_uri_to_url() -> None:
    mock_uri = "s3://bucket/data.csv"
    mock_local_uri = "s3://bucket/data.csv"
    assert (
        convert_uri_to_url(mock_uri, mock_s3_resource.endpoint_url)
        == f"https://{mock_bucket}.s3.{mock_region}.amazonaws.com/data.csv"
    )
    assert (
        convert_uri_to_url(mock_local_uri, mock_local_s3_resource.endpoint_url)
        == f"http://localhost:9000/{mock_bucket}/data.csv"
    )
    with pytest.raises(ValueError, match="endpoint_url must be provided"):
        convert_uri_to_url("", None)


def test_deferred_string() -> None:
    deferred = create_predicate("uses {context}")
    assert deferred(context="context_now") == "uses context_now"


def test_dataframe_to_list() -> None:
    generic_dataframe = DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    result = dataframe_to_list(generic_dataframe, dict)
    assert result == [{"A": 1, "B": 4}, {"A": 2, "B": 5}, {"A": 3, "B": 6}]

def test_list_to_dataframe() -> None:
    """Test that list_to_dataframe converts Pydantic models to DataFrame."""
    class TestModel(BaseModel):
        name: str
        value: int
    
    models = [
        TestModel(name="Item 1", value=10),
        TestModel(name="Item 2", value=20),
    ]
    
    result = list_to_dataframe(models)
    
    assert isinstance(result, DataFrame)
    assert len(result) == 2
    assert list(result.columns) == ["name", "value"]
    assert result.iloc[0]["name"] == "Item 1"
    assert result.iloc[0]["value"] == 10
    assert result.iloc[1]["name"] == "Item 2"
    assert result.iloc[1]["value"] == 20

def test_name_to_uuid() -> None:
    """Test that name_to_uuid produces consistent UUIDs."""
    name1 = "Test Entity"
    name2 = "test entity"
    name3 = "  Test Entity  "
    
    uuid1 = name_to_uuid(name1)
    uuid2 = name_to_uuid(name2)
    uuid3 = name_to_uuid(name3)
    
    assert isinstance(uuid1, UUID)
    assert uuid1 == uuid2
    assert uuid1 == uuid3

def test_uuid_from_s3_object() -> None:
    """Test that uuid_from_s3_object computes UUID from S3 object bytes."""
    # Create a mock S3Path
    mock_path = Mock(spec=S3Path)
    mock_path.read_bytes.return_value = b"test content"
    
    # Calculate expected UUID (name_to_uuid of hex of bytes)
    expected_hex = b"test content".hex()
    expected_uuid = name_to_uuid(expected_hex)
    
    result = uuid_from_s3_object(mock_path)
    
    assert isinstance(result, str)
    assert result == str(expected_uuid)
    mock_path.read_bytes.assert_called_once()


import os
from uuid import uuid4, UUID
import pytest
from horizon_data_core.client import PostgresClient
from horizon_data_core.sdk import HorizonSDK, IcebergRepository
from horizon_data_core.base_types import Entity, DataStream
from sqlalchemy import text
from unittest.mock import Mock


@pytest.fixture
def initialized_sdk() -> HorizonSDK:
    """Fixture to initialize SDK with database connection."""
    postgres_client = PostgresClient(
        user=os.environ.get("OWNER_PGUSER", "postgres"),
        password=os.environ.get("OWNER_PGPASSWORD", "password"),
        host=os.environ.get("PGHOST", "localhost"),
        port=int(os.environ.get("PGPORT", 5432)),
        database=os.environ.get("PGDATABASE", "horizon"),
        sslmode="disable",
        channel_binding="prefer",
    )
    iceberg_catalog_properties: dict[str, str | None] = {
        "uri": "http://localhost:8181",
        "warehouse": "file://./iceberg/warehouse",
    }
    return HorizonSDK(postgres_client, iceberg_catalog_properties, None)


@pytest.fixture
def entity_kind_id(initialized_sdk: HorizonSDK) -> UUID:
    """Fixture to create an entity kind for use in repository tests."""
    kind_id = uuid4()
    with initialized_sdk.postgres_client.session() as session:
        session.execute(
            text("INSERT INTO horizon_public.entity_kind (id, name) VALUES (:id, :name)"),
            {"id": kind_id, "name": "Test Entity Kind"}
        )
        session.commit()
    return kind_id


def test_create_sdk() -> None:
    """Test that the SDK can be created."""
    postgres_client = PostgresClient(
        user="postgres",
        password="password",
        host="localhost",
        port=5432,
        database="horizon",
        sslmode="disable",
        channel_binding="prefer",
    )
    iceberg_catalog_properties: dict[str, str | None] = {
        "uri": "http://localhost:8181",
        "warehouse": "file://./iceberg/warehouse",
    }
    sdk = HorizonSDK(postgres_client, iceberg_catalog_properties, None)
    assert sdk is not None


# Repository Operation Tests

def test_repository_insert(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository insert() creates a new record."""
    entity = Entity(
        id=uuid4(),
        kind_id=entity_kind_id,
        name="Test Entity",
    )
    result = initialized_sdk.entities.insert(entity)
    assert result.id == entity.id
    assert result.name == "Test Entity"
    assert result.kind_id == entity_kind_id


def test_repository_insert_batch(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository insert_batch() creates multiple records."""
    entities = [
        Entity(id=uuid4(), kind_id=entity_kind_id, name="Entity 1"),
        Entity(id=uuid4(), kind_id=entity_kind_id, name="Entity 2"),
    ]
    result = initialized_sdk.entities.insert_batch(entities)
    assert len(result) == 2
    assert all(r.name in ["Entity 1", "Entity 2"] for r in result)


def test_repository_insert_batch_empty(initialized_sdk: HorizonSDK) -> None:
    """Test that repository insert_batch() returns empty list for empty input."""
    result = initialized_sdk.entities.insert_batch([])
    assert result == []


def test_repository_read(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository read() retrieves a record by ID."""
    entity = Entity(id=uuid4(), kind_id=entity_kind_id, name="Test Entity")
    created = initialized_sdk.entities.insert(entity)
    assert created.id is not None
    
    result = initialized_sdk.entities.read(created.id)
    assert result is not None
    assert result.id == created.id
    assert result.name == "Test Entity"


def test_repository_read_not_found(initialized_sdk: HorizonSDK) -> None:
    """Test that repository read() returns None for non-existent ID."""
    result = initialized_sdk.entities.read(uuid4())
    assert result is None


def test_repository_update(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository update() modifies an existing record."""
    entity = Entity(id=uuid4(), kind_id=entity_kind_id, name="Original Name")
    created = initialized_sdk.entities.insert(entity)
    assert created.id is not None
    
    updated_entity = Entity(id=created.id, kind_id=entity_kind_id, name="Updated Name")
    result = initialized_sdk.entities.update(updated_entity)
    assert result.name == "Updated Name"
    
    # Verify update persisted
    read_result = initialized_sdk.entities.read(created.id)
    assert read_result is not None
    assert read_result.name == "Updated Name"


def test_repository_delete(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository delete() removes a record."""
    entity = Entity(id=uuid4(), kind_id=entity_kind_id, name="Entity to delete")
    created = initialized_sdk.entities.insert(entity)
    assert created.id is not None
    
    deleted = initialized_sdk.entities.delete(created.id)
    assert deleted is True
    
    # Verify it's gone
    result = initialized_sdk.entities.read(created.id)
    assert result is None


def test_repository_delete_not_found(initialized_sdk: HorizonSDK) -> None:
    """Test that repository delete() returns False for non-existent ID."""
    result = initialized_sdk.entities.delete(uuid4())
    assert result is False


def test_repository_list(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository list() returns all records."""
    entity1 = initialized_sdk.entities.insert(Entity(id=uuid4(), kind_id=entity_kind_id, name="Entity 1"))
    entity2 = initialized_sdk.entities.insert(Entity(id=uuid4(), kind_id=entity_kind_id, name="entity 2"))
    
    result = initialized_sdk.entities.list()
    assert len(result) >= 2
    entity_ids = {e.id for e in result}
    assert entity1.id in entity_ids
    assert entity2.id in entity_ids


def test_repository_list_with_filters(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository list() filters records correctly."""
    entity1 = initialized_sdk.entities.insert(Entity(id=uuid4(), kind_id=entity_kind_id, name="Filtered Entity"))
    initialized_sdk.entities.insert(Entity(id=uuid4(), kind_id=entity_kind_id, name="Other Entity"))
    
    # Filter by name
    result = initialized_sdk.entities.list(name="Filtered Entity")  # type: ignore[arg-type]
    assert len(result) >= 1
    assert all(e.name == "Filtered Entity" for e in result)
    assert entity1.id in {e.id for e in result}
    
    # Filter by kind_id
    kind_result = initialized_sdk.entities.list(kind_id=entity_kind_id)  # type: ignore[arg-type]
    assert len(kind_result) >= 2
    assert all(e.kind_id == entity_kind_id for e in kind_result)


def test_repository_upsert_create(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository upsert() creates a new record when it doesn't exist."""
    entity = Entity(id=uuid4(), kind_id=entity_kind_id, name="New Entity")
    result = initialized_sdk.entities.upsert(entity, unique_fields=["id"], merge_fields=["name", "kind_id"])
    assert result.id == entity.id
    assert result.id is not None
    
    # Verify it was created
    read_result = initialized_sdk.entities.read(result.id)
    assert read_result is not None
    assert read_result.name == "New Entity"


def test_repository_upsert_update(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository upsert() updates an existing record."""
    entity = Entity(id=uuid4(), kind_id=entity_kind_id, name="Original Name")
    created = initialized_sdk.entities.insert(entity)
    assert created.id is not None
    
    # Upsert with updated name
    updated = Entity(id=created.id, kind_id=entity_kind_id, name="Upserted Name")
    result = initialized_sdk.entities.upsert(updated, unique_fields=["id"], merge_fields=["name"])
    assert result.name == "Upserted Name"
    
    # Verify update persisted
    read_result = initialized_sdk.entities.read(created.id)
    assert read_result is not None
    assert read_result.name == "Upserted Name"


def test_repository_upsert_batch(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that repository upsert_batch() creates or updates multiple records."""
    # Create one entity first
    existing = initialized_sdk.entities.insert(Entity(id=uuid4(), kind_id=entity_kind_id, name="Existing"))
    assert existing.id is not None
    
    # Upsert batch with one existing and one new
    entities = [
        Entity(id=existing.id, kind_id=entity_kind_id, name="Updated Existing"),
        Entity(id=uuid4(), kind_id=entity_kind_id, name="New Entity"),
    ]
    result = initialized_sdk.entities.upsert_batch(entities, unique_fields=["id"], merge_fields=["name", "kind_id"])
    assert len(result) == 2
    
    # Verify existing was updated
    read_existing = initialized_sdk.entities.read(existing.id)
    assert read_existing is not None
    assert read_existing.name == "Updated Existing"
    
    # Verify new was created
    new_ids = {e.id for e in result}
    assert entities[1].id in new_ids

def test_sdk_create_data_stream_idempotent_id(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that creating data streams with the same platform and name produces the same ID."""
    # Create an entity to use as the platform for the data stream
    entity = Entity(
        id=uuid4(),
        kind_id=entity_kind_id,
        name="Test Platform Entity",
    )

    created_entity = initialized_sdk.entities.insert(entity)
    assert created_entity.id is not None
    data_stream_name = "Acoustic"

    # Create first data stream using from_platform_and_name
    data_stream_one = DataStream.from_platform_and_name(created_entity, data_stream_name)
    result_one = initialized_sdk.create_data_stream(data_stream_one)
    assert result_one.id is not None

    # Create second data stream with the same platform and name
    data_stream_two = DataStream.from_platform_and_name(created_entity, data_stream_name)
    result_two = initialized_sdk.create_data_stream(data_stream_two)
    assert result_two.id is not None

    # Verify both data streams have the same ID (idempotent)
    assert result_one.id == result_two.id

def test_sdk_read_data_stream(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK read_data_stream calls the repository read method."""
    initialized_sdk.data_streams.read = Mock()  # type: ignore[method-assign]
    initialized_sdk.read_data_stream(uuid4())
    initialized_sdk.data_streams.read.assert_called_once()

def test_sdk_update_data_stream(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK update_data_stream calls the repository upsert method."""
    initialized_sdk.data_streams.upsert = Mock()  # type: ignore[method-assign]
    initialized_sdk.update_data_stream(Mock())
    initialized_sdk.data_streams.upsert.assert_called_once()

def test_sdk_delete_data_stream(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK delete_data_stream calls the repository delete method."""
    initialized_sdk.data_streams.delete = Mock()  # type: ignore[method-assign]
    initialized_sdk.delete_data_stream(uuid4())
    initialized_sdk.data_streams.delete.assert_called_once()

def test_sdk_list_data_streams(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK list_data_streams calls the repository list method."""
    initialized_sdk.data_streams.list = Mock()  # type: ignore[method-assign]
    initialized_sdk.list_data_streams()
    initialized_sdk.data_streams.list.assert_called_once()

def test_sdk_create_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK create_mission calls the repository insert method."""
    initialized_sdk.missions.insert = Mock()  # type: ignore[method-assign]
    initialized_sdk.create_mission(Mock())
    initialized_sdk.missions.insert.assert_called_once()

def test_sdk_read_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK read_mission calls the repository read method."""
    initialized_sdk.missions.read = Mock()  # type: ignore[method-assign]
    initialized_sdk.read_mission(Mock())
    initialized_sdk.missions.read.assert_called_once()

def test_sdk_update_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK update_mission calls the repository update method."""
    initialized_sdk.missions.update = Mock()  # type: ignore[method-assign]
    initialized_sdk.update_mission(Mock())
    initialized_sdk.missions.update.assert_called_once()

def test_sdk_delete_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that SDK read_mission calls the repository delete method."""
    initialized_sdk.missions.delete = Mock()  # type: ignore[method-assign]
    initialized_sdk.delete_mission(Mock())
    initialized_sdk.missions.delete.assert_called_once()



"""Utilities for common project tasks."""

import re
from collections.abc import Callable, Iterable
from uuid import UUID, uuid5

from dagster import InputContext
from dagster_aws.s3 import S3Resource
from geojson import Feature, FeatureCollection, Point
from pandas import DataFrame, Series
from pydantic import BaseModel
from s3path import S3Path
from turfpy.measurement import center


def list_to_dataframe[T: BaseModel](object_iterable: Iterable[T]) -> DataFrame:
    """Coerces a list of Pydantic models to a pandas dataframe."""
    return DataFrame([object_row.model_dump() for object_row in object_iterable])


def dataframe_to_list[A](dataframe: DataFrame, cls: type[A]) -> list[A]:
    """Coerces a pandas dataframe into a list of Pydantic models."""
    instance_list: list[A] = []
    for _, row in dataframe.iterrows():
        instance = cls(**row.to_dict())
        instance_list.append(instance)
    return instance_list


namespace_uuid = UUID("13db1bb5-27ac-4bf7-8186-1043347c4247")


def name_to_uuid(name: str) -> UUID:
    """Produce consistent UUIDs for data idempotency."""
    return uuid5(namespace_uuid, name.strip().lower())


def uuid_from_s3_object(path: S3Path) -> str:
    """Compute the UUID of an S3 object given the S3Path to that object."""
    return str(name_to_uuid(path.read_bytes().hex()))


def get_center_of_points(points: list[Point] | Series) -> Point:  # type:ignore[no-any-unimported]
    """Get the center point of a list of geojson geometry points."""
    feature_point_list = [Feature(geometry=point) for point in points]
    return center(FeatureCollection(feature_point_list))["geometry"]


def create_predicate(template: str) -> Callable:
    """Create a function to delay evaluation of a f-string template.

    Create a function that takes `context` as input. This will be evaluated during
    `load_input` and that value of `context` will be within scope.

    Do not provide an f-string as input, as that will lead to immediate formatting. Instead
    provide a regular string with the same formatting syntax as an f-string.

    Example:
        >>> predicate = create_predicate("id='{context.partition_key}'")
        >>> lambda context: "id='{context.partition_key}'".format(context=context)

        Deferring evaluation to within the `load_input` scope results in a predicate clause
        of `"WHERE id='<context.partition_key value>'"`

    Returns:
        - Function that returns an f-string version of the input template, with the
    local context in scope.
    """

    def deferred_string_format(context: InputContext) -> str:
        """Evaluate a template string with the provided context value."""
        return template.format(context=context)

    return deferred_string_format


def transport_params(s3: S3Resource) -> dict:  # pragma: no cover
    """Create the transport_params inputs.

    Inteneded usage is func(inputs,**transport_params(s3))
    """
    return {"transport_params": {"client": s3.get_client()}}


def convert_uri_to_url(uri: str, endpoint_url: str | None) -> str:
    """Convert a URI to a URL."""
    if endpoint_url is None:
        message = "endpoint_url must be provided"
        raise ValueError(message)
    s3_path = S3Path.from_uri(uri)
    bucket = s3_path.bucket
    key = s3_path.key
    if endpoint_url and "amazonaws" in endpoint_url:
        service = "s3"
        region = ""
        if endpoint_url and "us-gov" in endpoint_url:
            pattern = r"s3(\.us-gov-[a-z0-9-]+)\.amazonaws\.com"
            match = re.search(pattern, endpoint_url)
            if match is not None:
                region = f"{match.group(1)}"
        url = f"https://{bucket}.{service}{region}.amazonaws.com/{key}"
    else:
        url = f"{endpoint_url}/{bucket}/{key}"
    return url



import os
import re
from typing import Any

import pytest
from dagster import build_input_context, build_output_context
from horizon_data_core.base_types import EntityKind
from horizon_data_core.helpers import create_predicate, list_to_dataframe, name_to_uuid
from horizon_data_core.resources.database import (
    PostgresSessionResource,
    construct_query_statement,
)
from horizon_data_core.resources.postgres_dataframe_io_manager import (
    PostgresDataframeIOManager,
    is_list_of_str_tuples,
)
from pandas import DataFrame
from sqlalchemy import text

postgres_kwargs: dict[str, Any] = {
    "username": os.environ.get("OWNER_PGUSER"),
    "password": os.environ.get("OWNER_PGPASSWORD"),
    "database_name": os.environ.get("PGDATABASE"),
    "hostname": os.environ.get("PGHOST", "localhost"),
    "port": int(os.environ.get("PGPORT", 5432)),
}


engine = PostgresSessionResource(**postgres_kwargs).get_engine()


def has_schema(schema_name: str) -> bool:
    session = PostgresSessionResource(**postgres_kwargs).get_session()
    res = session.execute(
        text("SELECT schema_name FROM information_schema.schemata WHERE schema_name = :schema_name"),
        {"schema_name": schema_name},
    )
    return res.fetchone() is not None


if has_schema("horizon_public"):
    SCHEMA_NAME = "horizon_public"
elif has_schema("underway_public"):
    SCHEMA_NAME = "underway_public"
else:
    raise ValueError("Expected schema horizon_public or underway_public to exist.")
print(f"Using schema: {SCHEMA_NAME}")


def remove_sql_formatting(string: str) -> str:
    """Replaces the following occurences with a single whitespace:
    - multiple consecutive whitespace
    - newline
    - tabs

    Also remove leading and trailing whitespace.
    """
    return re.sub(r"\s+", " ", string).strip(" ")


def mock_entity_kind_list() -> list[EntityKind]:
    return [
        EntityKind(
            id=name_to_uuid("Boeing Test-Frame"),
            name="Boeing Test-Frame",
            short_description="Test text.",
            long_description="A test set of longer text.",
            image_url=f"{name_to_uuid('Boeing Test-Frame')}.jpeg",
        ),
        EntityKind(
            id=name_to_uuid("SSN-99"),
            name="S",
            short_description="Test text.",
            long_description="A test set of longer text.",
            image_url=f"{name_to_uuid('SSN-99')}.jpeg",
        ),
    ]


def mock_entity_kind_dataframe() -> DataFrame:
    return list_to_dataframe(mock_entity_kind_list())


def test_cleanup_table() -> None:
    """Ensure we can delete records."""
    session = PostgresSessionResource(**postgres_kwargs).get_session()
    res = session.execute(text(f"DELETE FROM {SCHEMA_NAME}.entity_kind"))
    session.commit()
    assert res is not None


def test_postgres_io_workflow() -> None:
    """Test a stateful workflow:
    - 1. Ensure that a list of objects can be written to PostgreSQL.
    - 2. Ensure that the list of objects can be recovered from PostgreSQL.
    - 3. Using postgres_connection_resource, ensure we can run an arbitrary query on PostgreSQL and get the result back.
    - 4. Using PostgresDataframeIOManager, ensure we can run an arbitrary query on PostgreSQL and get the result back
        as a list of objects.
    """

    def test_postgres_io_manager_handle_output(
        io_manager: PostgresDataframeIOManager,
    ) -> None:
        """Check that the `PostgresIOManager` can insert the mock data as records in PostgreSQL."""
        entity_kind_context = build_output_context(
            metadata={
                "schema_name": SCHEMA_NAME,
                "table_name": "entity_kind",
                "should_upsert": True,
            },
        )
        io_manager.handle_output(
            entity_kind_context,
            mock_entity_kind_dataframe(),
        )

    def test_postgres_io_manager_handle_input(
        io_manager: PostgresDataframeIOManager,
    ) -> None:
        """Check that the pg io manager can use the load_input method to recover the list of objects."""
        context = build_input_context(
            metadata={
                "schema_name": SCHEMA_NAME,
                "table_name": "entity_kind",
            },
        )
        entity_kind_dataframe = io_manager.load_input(context)
        assert len(entity_kind_dataframe) == len(mock_entity_kind_dataframe())

    def test_postgres_io_manager_handle_input_from_output_context(
        io_manager: PostgresDataframeIOManager,
    ) -> None:
        """
        Check that the pg io manager can use the load_input method to recover the list of objects using the
        upstream_output metadata.
        """
        context = build_input_context(
            upstream_output=build_output_context(
                metadata={
                    "schema_name": SCHEMA_NAME,
                    "table_name": "entity_kind",
                },
            ),
        )
        entity_kind_dataframe = io_manager.load_input(context)
        assert len(entity_kind_dataframe) == len(mock_entity_kind_dataframe())

    def test_postgres_connection_resource_aggregation() -> None:
        """Use the postgres connection resource to run an arbitrary query."""
        session = PostgresSessionResource(**postgres_kwargs).get_session()
        res = session.execute(text(f"SELECT COUNT(*) FROM {SCHEMA_NAME}.entity_kind"))
        count = res.fetchall()[0][0]
        assert count == len(mock_entity_kind_dataframe())

    def test_postgres_io_manager_handle_queried_input() -> None:
        """Set the io manager to load data using an arbitary query and get the results back."""
        io_manager = PostgresDataframeIOManager(
            **postgres_kwargs,
        )
        entity_kind_id = mock_entity_kind_list()[1].id
        context = build_input_context(
            metadata={
                "schema_name": SCHEMA_NAME,
                "table_name": "entity_kind",
                "columns": ["id"],
                "predicate": create_predicate(f"id = '{entity_kind_id}'"),
            },
        )
        singleton_result_dataframe = io_manager.load_input(context)
        singleton_mock_dataframe = DataFrame(
            {"id": [f"{entity_kind_id}"]},
        )
        assert all(singleton_result_dataframe == singleton_mock_dataframe)

    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    test_postgres_io_manager_handle_output(io_manager)
    test_postgres_io_manager_handle_input(io_manager)
    test_postgres_io_manager_handle_input_from_output_context(io_manager)
    test_postgres_connection_resource_aggregation()
    test_postgres_io_manager_handle_queried_input()


def test_postgres_io_manager_handle_empty_output() -> None:
    """Test the situation where nothing is handed to the io manager."""
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    context = build_output_context(
        metadata={"schema_name": SCHEMA_NAME, "table_name": "entity_kind"},
    )
    io_manager.handle_output(context, None)


def test_construct_query_statement_default() -> None:
    """Ensure that the constructed default query is what we expect."""
    table = PostgresSessionResource(**postgres_kwargs).get_table(
        engine,
        SCHEMA_NAME,
        "entity_kind",
    )
    statement = construct_query_statement(table)
    expected = sorted(
        remove_sql_formatting(
            f"""
        SELECT
        {SCHEMA_NAME}.entity_kind.id,
        {SCHEMA_NAME}.entity_kind.created_datetime,
        {SCHEMA_NAME}.entity_kind.modified_datetime,
        {SCHEMA_NAME}.entity_kind.image_url,
        {SCHEMA_NAME}.entity_kind.name,
        {SCHEMA_NAME}.entity_kind.short_description,
        {SCHEMA_NAME}.entity_kind.long_description,
        {SCHEMA_NAME}.entity_kind.organization_id,
        {SCHEMA_NAME}.entity_kind.content_vector
        FROM
            {SCHEMA_NAME}.entity_kind
        """,
        ).split(),
    )
    received = sorted(remove_sql_formatting(str(statement)).split())
    message = f"Received: {received} does not match expected: {expected}."
    assert expected == received, message


def test_construct_query_statement_with_column_subsets_and_predicates() -> None:
    """Ensure that a query constructed with column subsets and predicates is what we expect."""
    entity_kind_id = mock_entity_kind_list()[0].id
    table = PostgresSessionResource(**postgres_kwargs).get_table(
        engine,
        SCHEMA_NAME,
        "entity_kind",
    )
    statement = construct_query_statement(
        table=table,
        query_column_list=["id"],
        query_predicate=f"id = '{entity_kind_id}'",
    )
    expected = remove_sql_formatting(
        f"""
    SELECT {SCHEMA_NAME}.entity_kind.id
    FROM {SCHEMA_NAME}.entity_kind
    WHERE id = '{entity_kind_id}'
    """,  # noqa: S608
    )
    received = remove_sql_formatting(str(statement))
    message = f"Recevied: {received} does not match expected: {expected}."
    assert expected == received, message


def test_construct_query_statement_with_join() -> None:
    entity_kind_id = mock_entity_kind_list()[0].id
    io_manager = PostgresDataframeIOManager(**postgres_kwargs)
    table = PostgresSessionResource(**postgres_kwargs).get_table(
        engine,
        SCHEMA_NAME,
        "entity",
    )
    statement = construct_query_statement(
        table=table,
        query_column_list=["id"],
        query_predicate=f"id = '{entity_kind_id}'",
        query_join=io_manager._get_join(  # noqa: SLF001
            {"join": [("entity_kind", "kind_id")], "schema_name": SCHEMA_NAME},
            io_manager._get_engine({}),  # noqa: SLF001
        ),
    )
    expected = remove_sql_formatting(
        f"""
    SELECT {SCHEMA_NAME}.entity.id
    FROM {SCHEMA_NAME}.entity
    JOIN {SCHEMA_NAME}.entity_kind ON {SCHEMA_NAME}.entity_kind.id = {SCHEMA_NAME}.entity.kind_id
    WHERE id = '{entity_kind_id}'
    """,  # noqa: S608
    )
    received = remove_sql_formatting(str(statement))
    message = f"Recevied: {received} does not match expected: {expected}."
    assert expected == received, message


def test_bad_construct_query_statement() -> None:
    """Ensure that the function raises an error if the desired column subset is incorrectly specified."""
    table = PostgresSessionResource(**postgres_kwargs).get_table(
        engine,
        SCHEMA_NAME,
        "entity_kind",
    )
    with pytest.raises(TypeError):
        construct_query_statement(
            table=table,
            query_column_list="id, date",  # type: ignore[arg-type]
        )


def test_postgres_io_manager_input_no_metadata() -> None:
    """Ensure that specifying no metadata throws an error."""
    context = build_input_context(
        metadata=None,
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    with pytest.raises(ValueError, match="upstream output or input metadata must be defined."):
        io_manager.load_input(context)


def test_postgres_io_manager_output_no_metadata() -> None:
    """Ensure that specifying no metadata throws an error."""
    context = build_output_context(
        metadata=None,
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    entity_kind_dataframe = mock_entity_kind_dataframe()
    with pytest.raises(ValueError, match="context.metadata cannot be `None`."):
        io_manager.handle_output(context, entity_kind_dataframe)


def test_postgres_io_manager_no_schema_name() -> None:
    """Ensure that specifying no schema name throws an error."""
    context = build_output_context(
        metadata={"table_name": "entity_kind"},
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    entity_kind_dataframe = mock_entity_kind_dataframe()
    with pytest.raises(ValueError, match="'schema_name' cannot be `None`."):
        io_manager.handle_output(context, entity_kind_dataframe)


def test_postgres_io_manager_no_table_name() -> None:
    context = build_output_context(
        metadata={"schema_name": SCHEMA_NAME},
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    entity_kind_dataframe = mock_entity_kind_dataframe()
    with pytest.raises(ValueError, match="'table_name' cannot be `None`."):
        io_manager.handle_output(context, entity_kind_dataframe)


def test_postgres_io_manager_nonstring_schema() -> None:
    int_schema = 1
    context = build_input_context(
        metadata={"schema_name": int_schema},
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    with pytest.raises(
        TypeError,
    ):
        io_manager.load_input(context)


def test_postgres_io_manager_nonstring_tablename() -> None:
    int_tablename = 1
    context = build_input_context(
        metadata={"schema_name": SCHEMA_NAME, "table_name": int_tablename},
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    with pytest.raises(
        TypeError,
    ):
        io_manager.load_input(context)


def test_postgres_io_manager_bad_join() -> None:
    test_cleanup_table()
    bad_join = [1, 2]
    context = build_input_context(
        metadata={
            "schema_name": SCHEMA_NAME,
            "table_name": "entity_kind",
            "join": bad_join,
        },
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    metadata = io_manager._load_input_context_metadata(context)  # noqa: SLF001
    engine = io_manager._get_engine(metadata)  # noqa: SLF001

    mock_entity_kind_dataframe()
    with pytest.raises(
        TypeError,
    ):
        io_manager._get_join(metadata, engine)  # noqa: SLF001


def test_postgres_io_manager_join() -> None:
    test_cleanup_table()
    join = [("entity_kind", "kind_id")]
    context = build_input_context(
        metadata={
            "schema_name": SCHEMA_NAME,
            "table_name": "entity_kind",
            "join": join,
        },
    )
    io_manager = PostgresDataframeIOManager(
        **postgres_kwargs,
    )
    metadata = io_manager._load_input_context_metadata(context)  # noqa: SLF001
    engine = io_manager._get_engine(metadata)  # noqa: SLF001

    mock_entity_kind_dataframe()
    io_manager._get_join(metadata, engine)  # noqa: SLF001


def test_is_list_of_str_tuples() -> None:
    assert is_list_of_str_tuples([1, 2]) is False
    assert is_list_of_str_tuples([("hello", "world"), ("entity_kind", "kind_id")]) is True




"""Public API for Horizon Data Core operations."""

import logging
from typing import Any
from uuid import UUID, uuid4

from .base_types import (
    BeamgramSpecification,
    BearingTimeRecordSpecification,
    DataRow,
    DataStream,
    Entity,
    EntityBeamgramSpecification,
    EntityBearingTimeRecordSpecification,
    MetadataRow,
    Mission,
    MissionEntity,
    Ontology,
    OntologyClass,
)
from .client import Client
from .sdk import HorizonSDK

# API scoped singleton SDK instance - will be initialized by users
_sdk: HorizonSDK | None = None


logger = logging.getLogger(__name__)


def initialize_sdk(
    client: Client,
    catalog_properties: dict[str, str | None],
    organization_id: UUID | None,
    iceberg_batch_size: int = 32,
    max_retries: int = 10,
) -> HorizonSDK:
    """Initialize the global SDK instance.

    This function sets up the global SDK instance that is used by all API functions.
    Must be called before using any other API functions.

    Args:
        client: PostgreSQL client for database operations
        iceberg_catalog: Iceberg catalog for table operations
        organization_id: Organization ID for multi-tenancy (can be None)
        iceberg_batch_size: Batch size for Iceberg operations (default: 32)
        max_retries: Maximum retry attempts for failed operations (default: 10)

    Raises:
        RuntimeError: If SDK is already initialized
    """
    # TODO(abuck): Refactor thehis to a singleton class instance pattern # noqa: FIX002
    # https://github.com/spear-ai/horizon/issues/998
    #  Alternatively, we could do away with the global SDK and just let users construct their own SDK instance
    #  and pass it around as needed.
    global _sdk  # noqa: PLW0603
    _sdk = HorizonSDK(client, catalog_properties, organization_id, iceberg_batch_size, max_retries)
    return _sdk


def get_horizon_sdk() -> HorizonSDK:
    """Get the global SDK instance."""
    return _ensure_sdk_initialized()


def _ensure_sdk_initialized() -> HorizonSDK:
    """Ensure the SDK is initialized."""
    if _sdk is None:
        message = "SDK not initialized. Call initialize_sdk() first."
        logger.error(message)
        raise RuntimeError(message)
    return _sdk


def create_entity_beamgram_specification(
    entity_id: UUID, beamgram_specification_id: UUID
) -> EntityBeamgramSpecification:
    """Create a new entity-beamgram specification relationship."""
    entity_beamgram_specification = EntityBeamgramSpecification(
        id=uuid4(), entity_id=entity_id, beamgram_specification_id=beamgram_specification_id
    )
    return _ensure_sdk_initialized().create_entity_beamgram_specification(entity_beamgram_specification)


def create_entity_bearing_time_record_specification(
    entity_id: UUID, bearing_time_record_specification_id: UUID
) -> EntityBearingTimeRecordSpecification:
    """Create a new entity-btr specification relationship."""
    entity_bearing_time_record_specification = EntityBearingTimeRecordSpecification(
        id=uuid4(), entity_id=entity_id, bearing_time_record_specification_id=bearing_time_record_specification_id
    )
    return _ensure_sdk_initialized().create_entity_bearing_time_record_specification(
        entity_bearing_time_record_specification
    )


# Entity operations
def create_entity(entity: Entity) -> Entity:
    """Create a new entity.

    Args:
        entity: The entity instance to create

    Returns:
        The created entity with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    logger.debug(
        "Creating entity",
        extra={"entity": entity},
    )
    return _ensure_sdk_initialized().create_entity(entity)


def upsert_entity(entity: Entity) -> Entity:
    """Create a new entity or update if it already exists.

    Args:
        entity: The entity instance to create or update

    Returns:
        The created or updated entity

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    logger.debug(
        "Creating or updating entity",
        extra={"entity": entity},
    )
    return _ensure_sdk_initialized().create_or_update_entity(entity)


def read_entity(id: UUID) -> Entity | None:
    """Read an entity by ID.

    Args:
        id: The UUID of the entity to read

    Returns:
        The entity if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_entity(id)


def update_entity(entity: Entity) -> Entity:
    """Update an existing entity.

    Args:
        entity: The entity instance with updated fields

    Returns:
        The updated entity

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If entity with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_entity(entity)


def delete_entity(id: UUID) -> bool:
    """Delete an entity by ID.

    Args:
        id: The UUID of the entity to delete

    Returns:
        True if the entity was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_entity(id)


def list_entities(**filters: Any) -> list[Entity]:  # noqa: ANN401
    """List entities with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of entities matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_entities(**filters)


# DataStream operations
def create_data_stream(platform: Entity, data_stream_name: str) -> DataStream:
    """Create a new data stream, or return the existing matching one.

    Args:
        platform: The platform this data_stream belongs to
        data_stream_name: The name of the stream or sensor

    Returns:
        The created data stream with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    data_stream = DataStream.from_platform_and_name(platform, data_stream_name)
    return _ensure_sdk_initialized().create_data_stream(data_stream)


def read_data_stream(id: UUID) -> DataStream | None:
    """Read a data stream by ID.

    Args:
        id: The UUID of the data stream to read

    Returns:
        The data stream if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_data_stream(id)


def update_data_stream(data_stream: DataStream) -> DataStream:
    """Update an existing data stream.

    Args:
        data_stream: The data stream instance with updated fields

    Returns:
        The updated data stream

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If data stream with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_data_stream(data_stream)


def delete_data_stream(id: UUID) -> bool:
    """Delete a data stream by ID.

    Args:
        id: The UUID of the data stream to delete

    Returns:
        True if the data stream was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_data_stream(id)


def list_data_streams(**filters: Any) -> list[DataStream]:  # noqa: ANN401
    """List data streams with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of data streams matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_data_streams(**filters)


# Mission operations
def create_mission(mission: Mission) -> Mission:
    """Create a new mission.

    Args:
        mission: The mission instance to create

    Returns:
        The created mission with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_mission(mission)


def read_mission(id: UUID) -> Mission | None:
    """Read a mission by ID.

    Args:
        id: The UUID of the mission to read

    Returns:
        The mission if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_mission(id)


def update_mission(mission: Mission) -> Mission:
    """Update an existing mission.

    Args:
        mission: The mission instance with updated fields

    Returns:
        The updated mission

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If mission with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_mission(mission)


def delete_mission(id: UUID) -> bool:
    """Delete a mission by ID.

    Args:
        id: The UUID of the mission to delete

    Returns:
        True if the mission was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_mission(id)


def list_missions(**filters: Any) -> list[Mission]:  # noqa: ANN401
    """List missions with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of missions matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_missions(**filters)


# MissionEntity operations
def create_mission_entity(mission_entity: MissionEntity) -> MissionEntity:
    """Create a new mission-entity relationship.

    Args:
        mission_entity: The mission-entity relationship instance to create

    Returns:
        The created mission-entity relationship with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_mission_entity(mission_entity)


def read_mission_entity(id: UUID) -> MissionEntity | None:
    """Read a mission-entity relationship by ID.

    Args:
        id: The UUID of the mission-entity relationship to read

    Returns:
        The mission-entity relationship if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_mission_entity(id)


def update_mission_entity(mission_entity: MissionEntity) -> MissionEntity:
    """Update an existing mission-entity relationship.

    Args:
        mission_entity: The mission-entity relationship instance with updated fields

    Returns:
        The updated mission-entity relationship

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If mission-entity relationship with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_mission_entity(mission_entity)


def delete_mission_entity(id: UUID) -> bool:
    """Delete a mission-entity relationship by ID.

    Args:
        id: The UUID of the mission-entity relationship to delete

    Returns:
        True if the mission-entity relationship was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_mission_entity(id)


def list_mission_entities(**filters: Any) -> list[MissionEntity]:  # noqa: ANN401
    """List mission-entity relationships with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of mission-entity relationships matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_mission_entities(**filters)


# Ontology operations
def create_ontology(ontology: Ontology) -> Ontology:
    """Create a new ontology.

    Args:
        ontology: The ontology instance to create

    Returns:
        The created ontology with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_ontology(ontology)


def read_ontology(id: UUID) -> Ontology | None:
    """Read an ontology by ID.

    Args:
        id: The UUID of the ontology to read

    Returns:
        The ontology if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_ontology(id)


def update_ontology(ontology: Ontology) -> Ontology:
    """Update an existing ontology.

    Args:
        ontology: The ontology instance with updated fields

    Returns:
        The updated ontology

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If ontology with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_ontology(ontology)


def delete_ontology(id: UUID) -> bool:
    """Delete an ontology by ID.

    Args:
        id: The UUID of the ontology to delete

    Returns:
        True if the ontology was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_ontology(id)


def list_ontologies(**filters: Any) -> list[Ontology]:  # noqa: ANN401
    """List ontologies with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of ontologies matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_ontologies(**filters)


# OntologyClass operations
def create_ontology_class(ontology_class: OntologyClass) -> OntologyClass:
    """Create a new ontology class.

    Args:
        ontology_class: The ontology class instance to create

    Returns:
        The created ontology class with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_ontology_class(ontology_class)


def read_ontology_class(id: UUID) -> OntologyClass | None:
    """Read an ontology class by ID.

    Args:
        id: The UUID of the ontology class to read

    Returns:
        The ontology class if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_ontology_class(id)


def update_ontology_class(ontology_class: OntologyClass) -> OntologyClass:
    """Update an existing ontology class.

    Args:
        ontology_class: The ontology class instance with updated fields

    Returns:
        The updated ontology class

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If ontology class with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_ontology_class(ontology_class)


def delete_ontology_class(id: UUID) -> bool:
    """Delete an ontology class by ID.

    Args:
        id: The UUID of the ontology class to delete

    Returns:
        True if the ontology class was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_ontology_class(id)


def list_ontology_classes(**filters: Any) -> list[OntologyClass]:  # noqa: ANN401
    """List ontology classes with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of ontology classes matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_ontology_classes(**filters)


# BeamgramSpecification operations
def create_beamgram_specification(beamgram_specification: BeamgramSpecification) -> BeamgramSpecification:
    """Create a new beamgram specification.

    Args:
        beamgram_specification: The beamgram specification instance to create

    Returns:
        The created beamgram specification with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_beamgram_specification(beamgram_specification)


def create_or_update_beamgram_specification(beamgram_specification: BeamgramSpecification) -> BeamgramSpecification:
    """Create a new beamgram specification or update if it already exists.

    Args:
        beamgram_specification: The beamgram specification instance to create or update

    Returns:
        The created or updated beamgram specification

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_or_update_beamgram_specification(beamgram_specification)


def read_beamgram_specification(id: UUID) -> BeamgramSpecification | None:
    """Read a beamgram specification by ID.

    Args:
        id: The UUID of the beamgram specification to read

    Returns:
        The beamgram specification if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_beamgram_specification(id)


def update_beamgram_specification(beamgram_specification: BeamgramSpecification) -> BeamgramSpecification:
    """Update an existing beamgram specification.

    Args:
        beamgram_specification: The beamgram specification instance with updated fields

    Returns:
        The updated beamgram specification

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If beamgram specification with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_beamgram_specification(beamgram_specification)


def delete_beamgram_specification(id: UUID) -> bool:
    """Delete a beamgram specification by ID.

    Args:
        id: The UUID of the beamgram specification to delete

    Returns:
        True if the beamgram specification was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_beamgram_specification(id)


def list_beamgram_specifications(**filters: Any) -> list[BeamgramSpecification]:  # noqa: ANN401
    """List beamgram specifications with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of beamgram specifications matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_beamgram_specifications(**filters)


# BearingTimeRecordSpecification operations
def create_bearing_time_record_specification(
    bearing_time_record_specification: BearingTimeRecordSpecification,
) -> BearingTimeRecordSpecification:
    """Create a new bearing-time record specification.

    Args:
        bearing_time_record_specification: The bearing-time record specification instance to create

    Returns:
        The created bearing-time record specification with updated fields (e.g., generated ID)

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_bearing_time_record_specification(bearing_time_record_specification)


def create_or_update_bearing_time_record_specification(
    bearing_time_record_specification: BearingTimeRecordSpecification,
) -> BearingTimeRecordSpecification:
    """Create a new bearing-time record specification or update if it already exists.

    Args:
        bearing_time_record_specification: The bearing-time record specification instance to create or update

    Returns:
        The created or updated bearing-time record specification

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().create_or_update_bearing_time_record_specification(
        bearing_time_record_specification
    )


def read_bearing_time_record_specification(id: UUID) -> BearingTimeRecordSpecification | None:
    """Read a bearing-time record specification by ID.

    Args:
        id: The UUID of the bearing-time record specification to read

    Returns:
        The bearing-time record specification if found, None otherwise

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().read_bearing_time_record_specification(id)


def update_bearing_time_record_specification(
    bearing_time_record_specification: BearingTimeRecordSpecification,
) -> BearingTimeRecordSpecification:
    """Update an existing bearing-time record specification.

    Args:
        bearing_time_record_specification: The bearing-time record specification instance with updated fields

    Returns:
        The updated bearing-time record specification

    Raises:
        RuntimeError: If SDK is not initialized
        ValueError: If bearing-time record specification with the given ID is not found
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().update_bearing_time_record_specification(bearing_time_record_specification)


def delete_bearing_time_record_specification(id: UUID) -> bool:
    """Delete a bearing-time record specification by ID.

    Args:
        id: The UUID of the bearing-time record specification to delete

    Returns:
        True if the bearing-time record specification was deleted, False if it didn't exist

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().delete_bearing_time_record_specification(id)


def list_bearing_time_record_specifications(**filters: Any) -> list[BearingTimeRecordSpecification]:  # noqa: ANN401
    """List bearing-time record specifications with optional filters.

    Args:
        **filters: Keyword arguments where keys are field names and values are filter values

    Returns:
        List of bearing-time record specifications matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        SQLAlchemyError: If database operation fails
    """
    return _ensure_sdk_initialized().list_bearing_time_record_specifications(**filters)


# Iceberg operations
def create_data_row(data_row: DataRow, buffer_name: str | None = None) -> DataRow:
    """Create a new data row in Iceberg table.

    Args:
        data_row: The data row instance to create
        buffer_name: Optional buffer name for batching writes

    Returns:
        The data row instance (unchanged)

    Raises:
        RuntimeError: If SDK is not initialized
        CommitFailedException: If write operation fails after retries
    """
    return _ensure_sdk_initialized().create_data_row(data_row, buffer_name)


def create_data_rows(data_rows: list[DataRow], buffer_name: str | None = None) -> list[DataRow]:
    """Create a new data rows in Iceberg table.

    Args:
        data_rows: The list of data row instances to create
        buffer_name: Optional buffer name for batching writes

    Returns:
        The list of data row instances (unchanged)
    """
    return _ensure_sdk_initialized().create_data_row_batch(data_rows, buffer_name)


def list_data_rows(**filters: Any) -> list[DataRow]:  # noqa: ANN401
    """List data rows with optional filters from Iceberg table.

    Args:
        **filters: Iceberg scan parameters including:
            - row_filter: str | boolean_expr = True
            - selected_fields: tuple[str] = ("*",)
            - case_sensitive: bool = True
            - snapshot_id: str | None = None
            - options: Properties = {}
            - limit: int | None = None

    Returns:
        List of data rows matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        Exception: If Iceberg scan operation fails
    """
    return _ensure_sdk_initialized().list_data_rows(**filters)


def create_metadata_row(metadata_row: MetadataRow, buffer_name: str | None = None) -> MetadataRow:
    """Create a new metadata row in Iceberg table.

    Args:
        metadata_row: The metadata row instance to create
        buffer_name: Optional buffer name for batching writes

    Returns:
        The metadata row instance (unchanged)

    Raises:
        RuntimeError: If SDK is not initialized
        CommitFailedException: If write operation fails after retries
    """
    return _ensure_sdk_initialized().create_metadata_row(metadata_row, buffer_name)


def list_metadata_rows(**filters: Any) -> list[MetadataRow]:  # noqa: ANN401
    """List metadata rows with optional filters from Iceberg table.

    Args:
        **filters: Iceberg scan parameters including:
            - row_filter: str | boolean_expr = True
            - selected_fields: tuple[str] = ("*",)
            - case_sensitive: bool = True
            - snapshot_id: str | None = None
            - options: Properties = {}
            - limit: int | None = None

    Returns:
        List of metadata rows matching the filters

    Raises:
        RuntimeError: If SDK is not initialized
        Exception: If Iceberg scan operation fails
    """
    return _ensure_sdk_initialized().list_metadata_rows(**filters)



import os
from uuid import uuid4, UUID
from typing import Generator
import pytest
from datetime import datetime
from horizon_data_core.api import (
    create_entity_beamgram_specification,
    create_entity_bearing_time_record_specification,
    get_horizon_sdk,
    initialize_sdk,
    create_beamgram_specification,
    create_bearing_time_record_specification,
    create_entity,
    read_entity,
    update_entity,
    delete_entity,
    list_entities,
    upsert_entity,
    create_data_rows,
    create_data_stream,
    read_data_stream,
    update_data_stream,
    list_data_streams,
    create_mission,
    read_mission,
    update_mission,
    delete_mission,
    list_missions,
    create_mission_entity,
    read_mission_entity,
    update_mission_entity,
    delete_mission_entity,
    list_mission_entities,
)
from horizon_data_core.client import PostgresClient
from horizon_data_core.base_types import Entity, BeamgramSpecification, BearingTimeRecordSpecification, DataRow, DataStream
from sqlalchemy import text
import horizon_data_core.api as api_module
from horizon_data_core.sdk import HorizonSDK
from unittest.mock import Mock, patch


@pytest.fixture(autouse=True)
def reset_sdk() -> Generator[None, None, None]:
    """Reset the global SDK before and after each test."""
    # Reset before test
    api_module._sdk = None
    yield
    # Reset after test
    api_module._sdk = None


@pytest.fixture
def initialized_sdk() -> HorizonSDK:
    """Fixture to initialize SDK with database connection for tests that need it."""
    initialize_sdk(
        client=PostgresClient(
            user=os.environ.get("OWNER_PGUSER", "postgres"),
            password=os.environ.get("OWNER_PGPASSWORD", "password"),
            host=os.environ.get("PGHOST", "localhost"),
            port=int(os.environ.get("PGPORT", 5432)),
            database=os.environ.get("PGDATABASE", "horizon"),
            sslmode="disable",
            channel_binding="prefer",
        ),
        catalog_properties={
            "uri": "http://localhost:8181",
            "warehouse": "file://./iceberg/warehouse",
        },
        organization_id=None,
    )
    return get_horizon_sdk()


@pytest.fixture
def entity_kind_id(initialized_sdk: HorizonSDK) -> UUID:
    """Fixture to create an entity kind for use in entity tests."""
    kind_id = uuid4()
    with initialized_sdk.postgres_client.session() as session:
        session.execute(
            text("INSERT INTO horizon_public.entity_kind (id, name) VALUES (:id, :name)"),
            {"id": kind_id, "name": "Test Entity Kind"}
        )
        session.commit()
    return kind_id


def test_initialize_sdk() -> None:
    """Test that the SDK can be initialized."""
    with pytest.raises(RuntimeError):
        get_horizon_sdk()
    sdk = initialize_sdk(
        client=PostgresClient(
            # Because this is a lazy connection we don't need to use real values.
            user="postgres",
            password="password",
            host="localhost",
            port=5432,
            database="horizon",
        ),
        catalog_properties={
            "uri": "http://localhost:8181",
            "warehouse": "file://./iceberg/warehouse",
        },
        organization_id=None,
    )
    assert sdk is not None
    sdk = get_horizon_sdk()
    assert sdk is not None


def test_create_entity_beamgram_specification(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that the entity beamgram specification can be created."""
    # Create a beamgram specification
    beamgram_spec = create_beamgram_specification(BeamgramSpecification(id=uuid4()))
    beamgram_id = beamgram_spec.id
    assert beamgram_id is not None
    
    # Create an entity with the valid kind_id
    entity = create_entity(Entity(id=uuid4(), kind_id=entity_kind_id, name="Test Entity"))
    entity_id = entity.id
    assert entity_id is not None

    spec = create_entity_beamgram_specification(entity_id, beamgram_id)
    assert spec is not None
    assert spec.entity_id == entity_id
    assert spec.beamgram_specification_id == beamgram_id


def test_create_entity_bearing_time_record_specification(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that the entity bearing time record specification can be created."""
    # Create a bearing time record specification
    btr_spec = create_bearing_time_record_specification(BearingTimeRecordSpecification(id=uuid4()))
    btr_id = btr_spec.id
    assert btr_id is not None
    
    # Create an entity with the valid kind_id
    entity = create_entity(Entity(id=uuid4(), kind_id=entity_kind_id, name="Test Entity"))
    entity_id = entity.id
    assert entity_id is not None

    spec = create_entity_bearing_time_record_specification(entity_id, btr_id)
    assert spec is not None
    assert spec.entity_id == entity_id
    assert spec.bearing_time_record_specification_id == btr_id


# Entity CRUD Tests

def test_create_entity(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API create_entity calls the SDK create_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_entity = Mock()  # type: ignore[method-assign]
        create_entity(Entity(id=uuid4(), kind_id=entity_kind_id, name="Test Entity"))
        initialized_sdk.create_entity.assert_called_once()


def test_read_entity(initialized_sdk: HorizonSDK) -> None:
    """Test that API read_entity calls the SDK read_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.read_entity = Mock()  # type: ignore[method-assign]
        read_entity(uuid4())
        initialized_sdk.read_entity.assert_called_once()


def test_read_entity_not_found(initialized_sdk: HorizonSDK) -> None:
    """Test that reading a non-existent entity returns None."""
    non_existent_id = uuid4()
    result = read_entity(non_existent_id)
    assert result is None


def test_update_entity(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API update_entity calls the SDK update_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.update_entity = Mock()  # type: ignore[method-assign]
        update_entity(Entity(id=uuid4(), kind_id=entity_kind_id, name="Test"))
        initialized_sdk.update_entity.assert_called_once()


def test_update_entity_not_found(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that updating a non-existent entity raises ValueError."""
    non_existent_entity = Entity(
        id=uuid4(),
        kind_id=entity_kind_id,
        name="Non-existent",
    )
    with pytest.raises(ValueError):
        update_entity(non_existent_entity)


def test_delete_entity(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API delete_entity calls the SDK delete_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.delete_entity = Mock()  # type: ignore[method-assign]
        delete_entity(uuid4())
        initialized_sdk.delete_entity.assert_called_once()


def test_delete_entity_not_found(initialized_sdk: HorizonSDK) -> None:
    """Test that deleting a non-existent entity returns False."""
    non_existent_id = uuid4()
    result = delete_entity(non_existent_id)
    assert result is False


def test_list_entities(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API list_entities calls the SDK list_entities method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.list_entities = Mock()  # type: ignore[method-assign]
        list_entities()
        initialized_sdk.list_entities.assert_called_once()


def test_list_entities_with_filters(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API list_entities calls the SDK list_entities method with filters."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.list_entities = Mock()  # type: ignore[method-assign]
        list_entities(name="Filtered Entity")
        initialized_sdk.list_entities.assert_called_once_with(name="Filtered Entity")


def test_upsert_entity_create(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API upsert_entity calls the SDK create_or_update_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_or_update_entity = Mock()  # type: ignore[method-assign]
        upsert_entity(Entity(id=uuid4(), kind_id=entity_kind_id, name="Upserted Entity"))
        initialized_sdk.create_or_update_entity.assert_called_once()


def test_upsert_entity_update(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API upsert_entity calls the SDK create_or_update_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_or_update_entity = Mock()  # type: ignore[method-assign]
        upsert_entity(Entity(id=uuid4(), kind_id=entity_kind_id, name="Upserted Name"))
        initialized_sdk.create_or_update_entity.assert_called_once()


def test_entity_operations_without_sdk() -> None:
    """Test that entity operations raise RuntimeError when SDK is not initialized."""
    entity_id = uuid4()
    entity = Entity(id=entity_id, kind_id=uuid4(), name="Test")
    
    with pytest.raises(RuntimeError):
        read_entity(entity_id)
    
    with pytest.raises(RuntimeError):
        update_entity(entity)
    
    with pytest.raises(RuntimeError):
        delete_entity(entity_id)
    
    with pytest.raises(RuntimeError):
        list_entities()
    
    with pytest.raises(RuntimeError):
        upsert_entity(entity)


def test_create_data_rows(initialized_sdk: HorizonSDK, entity_kind_id: UUID) -> None:
    """Test that API create_data_rows calls the SDK create_data_row_batch method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_data_row_batch = Mock()  # type: ignore[method-assign]
        create_data_rows([Mock()])
        initialized_sdk.create_data_row_batch.assert_called_once()

def test_create_data_stream(initialized_sdk: HorizonSDK) -> None:
    """Test that API create_data_stream calls the SDK create_data_stream method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_data_stream = Mock()  # type: ignore[method-assign]
        mock_platform = Mock()
        mock_platform.id = uuid4()
        create_data_stream(mock_platform, "TestDataStream")
        initialized_sdk.create_data_stream.assert_called_once()

def test_read_data_stream(initialized_sdk: HorizonSDK) -> None:
    """Test that API read_data_stream calls the SDK read_data_stream method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.read_data_stream = Mock()  # type: ignore[method-assign]
        read_data_stream(uuid4())
        initialized_sdk.read_data_stream.assert_called_once()

def test_update_data_stream(initialized_sdk: HorizonSDK) -> None:
    """Test that API update_data_stream calls the SDK update_data_stream method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.update_data_stream = Mock()  # type: ignore[method-assign]
        update_data_stream(Mock())
        initialized_sdk.update_data_stream.assert_called_once()

def test_list_data_streams(initialized_sdk: HorizonSDK) -> None:
    """Test that API list_data_streams calls the SDK list_data_streams method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.list_data_streams = Mock()  # type: ignore[method-assign]
        list_data_streams()
        initialized_sdk.list_data_streams.assert_called_once()

def test_create_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that API create_mission calls the SDK create_mission method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_mission = Mock()  # type: ignore[method-assign]
        create_mission(Mock())
        initialized_sdk.create_mission.assert_called_once()

def test_read_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that API read_mission calls the SDK read_mission method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.read_mission = Mock()  # type: ignore[method-assign]
        read_mission(uuid4())
        initialized_sdk.read_mission.assert_called_once()

def test_update_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that API update_mission calls the SDK update_mission method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.update_mission = Mock()  # type: ignore[method-assign]
        update_mission(Mock())
        initialized_sdk.update_mission.assert_called_once()

def test_delete_mission(initialized_sdk: HorizonSDK) -> None:
    """Test that API delete_mission calls the SDK delete_mission method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.delete_mission = Mock()  # type: ignore[method-assign]
        delete_mission(uuid4())
        initialized_sdk.delete_mission.assert_called_once()

def test_list_missions(initialized_sdk: HorizonSDK) -> None:
    """Test that API list_missions calls the SDK list_missions method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.list_missions = Mock()  # type: ignore[method-assign]
        list_missions()
        initialized_sdk.list_missions.assert_called_once()

def test_create_mission_entity(initialized_sdk: HorizonSDK) -> None:
    """Test that API create_mission_entity calls the SDK create_mission_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.create_mission_entity = Mock()  # type: ignore[method-assign]
        create_mission_entity(Mock())
        initialized_sdk.create_mission_entity.assert_called_once()

def test_read_mission_entity(initialized_sdk: HorizonSDK) -> None:
    """Test that API read_mission_entity calls the SDK read_mission_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.read_mission_entity = Mock()  # type: ignore[method-assign]
        read_mission_entity(uuid4())
        initialized_sdk.read_mission_entity.assert_called_once()

def test_update_mission_entity(initialized_sdk: HorizonSDK) -> None:
    """Test that API update_mission_entity calls the SDK update_mission_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.update_mission_entity = Mock()  # type: ignore[method-assign]
        update_mission_entity(Mock())
        initialized_sdk.update_mission_entity.assert_called_once()

def test_delete_mission_entity(initialized_sdk: HorizonSDK) -> None:
    """Test that API delete_mission_entity calls the SDK delete_mission_entity method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.delete_mission_entity = Mock()  # type: ignore[method-assign]
        delete_mission_entity(uuid4())
        initialized_sdk.delete_mission_entity.assert_called_once()

def test_list_mission_entities(initialized_sdk: HorizonSDK) -> None:
    """Test that API list_mission_entities calls the SDK list_mission_entities method."""
    with patch('horizon_data_core.api.get_horizon_sdk', return_value=initialized_sdk):
        initialized_sdk.list_mission_entities = Mock()  # type: ignore[method-assign]
        list_mission_entities()
        initialized_sdk.list_mission_entities.assert_called_once()



"""Logging setup for Horizon Data Pipelines."""

import functools
import logging
import sys
import time
from collections.abc import Callable
from typing import Final, ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")


def log_calls(logger: logging.Logger, level: int = logging.INFO) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Log function entry and exit with a correlation ID.

    Args:
        logger: The logger instance to use
        level: Logging level (e.g., logging.INFO, logging.DEBUG)

    Example:
        >>> logger = logging.getLogger(__name__)
        >>> @log_calls(logger, logging.DEBUG)
        ... def my_function(x: int, y: int) -> int:
        ...     return x + y
    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            # Use timestamp as correlation ID
            call_id = int(time.time() * 1_000_000) % 1_000_000_000

            # Log entry - let logger do the formatting
            logger.log(level, "→ ENTER [%s] %s", call_id, func.__name__)

            try:
                result = func(*args, **kwargs)
            except Exception as e:
                # Log exception exit
                logger.log(level, "✗ ERROR [%s] %s raised %s: %s", call_id, func.__name__, type(e).__name__, e)
                raise
            else:
                # Log successful exit
                logger.log(level, "← EXIT  [%s] %s", call_id, func.__name__)
                return result

        return wrapper

    return decorator


def setup_extra_logger() -> logging.Logger:
    """Set up the root logger for extra fields."""
    # Create a handler for the standard output
    hnd = logging.StreamHandler(sys.stdout)
    # Set the formatter to our formatter than handles "extra" fields
    hnd.setFormatter(ExFormatter("%(name)s - %(levelname)s - %(message)s"))
    # Get the root logger
    logger = logging.getLogger()
    # Clear any existing handlers to avoid duplicates on stdout
    logger.handlers.clear()
    # Add the new handler to the root logger
    logger.addHandler(hnd)
    return logger


__all__ = ["setup_extra_logger"]


class ExFormatter(logging.Formatter):
    def_keys: Final[list[str]] = [
        "name",
        "msg",
        "args",
        "levelname",
        "levelno",
        "pathname",
        "filename",
        "module",
        "exc_info",
        "exc_text",
        "stack_info",
        "lineno",
        "funcName",
        "created",
        "msecs",
        "relativeCreated",
        "thread",
        "threadName",
        "processName",
        "process",
        "message",
    ]

    def format(self, record: logging.LogRecord) -> str:
        """Format a log record."""
        string = super().format(record)
        extra = {k: v for k, v in record.__dict__.items() if k not in self.def_keys}
        if len(extra) > 0:
            string += " - extra: " + str(extra)
        return string



"""Common S3 based resources for any data pipeline."""

from typing import Self

from dagster import ConfigurableResource
from s3path import S3Path


class S3Bucket(ConfigurableResource):
    """Configuration for the data ingest bucket."""

    s3_bucket: str
    s3_prefix: str = ""

    def path(self: Self) -> S3Path:
        """Return the S3Path for the configured bucket and prefix."""
        return S3Path(f"/{self.s3_bucket}/{self.s3_prefix}")



"""Database resources for use in Dagster."""

from collections.abc import Callable, Iterable
from typing import Any

from dagster import ConfigurableResource
from pandas.io.sql import SQLTable
from sqlalchemy import Connection, Engine, MetaData, Select, Table, create_engine, select, text
from sqlalchemy.orm import Session, sessionmaker


class PostgresSessionResource(ConfigurableResource):
    """Resource for creating a Postgress session."""

    username: str = "postgres"
    password: str = "postgres"  # noqa: S105
    hostname: str = "localhost"
    port: int = 5432
    database_name: str = "postgres"
    ssl_mode: str | None = None

    def _get_connection_str(self: "PostgresSessionResource") -> str:
        """Return a Postgres session for Dagster."""
        return f"postgresql://{self.username}:{self.password}@{self.hostname}:{self.port}/{self.database_name}"

    def get_engine(self: "PostgresSessionResource") -> Engine:
        """Return an engine for accessing Postgres with Dagster."""
        return create_engine(
            self._get_connection_str(),
            connect_args={"sslmode": self.ssl_mode},
        )

    def get_session(self: "PostgresSessionResource") -> Session:
        """Return a Postgres session for Dagster."""
        return sessionmaker(bind=self.get_engine())()

    def get_table(
        self: "PostgresSessionResource",
        engine: Engine,
        schema_name: str,
        table_name: str,
    ) -> Table:
        """Acquire a table type definition for query statements."""
        metadata = MetaData(schema=schema_name)
        return Table(table_name, metadata, autoload_with=engine)


def construct_query_statement(
    table: Table,
    query_column_list: list[str] | None = None,
    query_join: list[tuple[Table, str]] | None = None,
    query_predicate: str | None = None,
) -> Select:
    """Return a sqlalchemy query statement."""
    if not query_column_list:
        statement = select(table)
    else:
        if not isinstance(query_column_list, list):
            msg = f"Expected query_column_list to be of type list. Got {type(query_column_list)} instead."
            raise TypeError(msg)
        columns = [table.c[col] for col in query_column_list]
        statement = select(*columns)
    if query_join:
        for join, on in query_join:
            statement = statement.join(join, join.c.id == table.c[on])
    if query_predicate:
        statement = statement.where(text(query_predicate))
    return statement


def create_upsert_method(
    schema_name: str,
    table_name: str,
) -> Callable[[SQLTable, Connection, list[str], Iterable[tuple[Any, ...]]], None]:
    """Return an upsert method for a specific schema and table."""

    def upsert(
        _table: SQLTable,
        conn: Connection,
        keys: list[str],
        data_iterator: Iterable[tuple[Any, ...]],
    ) -> None:
        """Perform an upsert compatible with Pandas's expected method function structure."""
        columns: str = ", ".join([f'"{k}"' for k in keys])
        insert_values: str = ", ".join([f":{k}" for k in keys])
        update_values: str = ", ".join([f'"{k}" = EXCLUDED."{k}"' for k in keys if k != "id"])

        upsert_query: str = f"""
            INSERT INTO {schema_name}.{table_name} ({columns})
            VALUES ({insert_values})
            ON CONFLICT (id)
            DO UPDATE SET {update_values}
        """  # noqa: S608

        prepared_statement = text(upsert_query)

        for data_chunk in data_iterator:
            statement_parameters = dict(zip(keys, data_chunk, strict=True))
            conn.execute(prepared_statement, statement_parameters)

    return upsert



from horizon_data_core.resources.s3 import S3Bucket
def test_s3_resource() -> None:
    """Test that the S3 resource can be created."""
    s3_resource = S3Bucket(
        s3_bucket="bucket",
        s3_prefix="prefix",
    )
    assert s3_resource.path().as_uri() == "s3://bucket/prefix"



import logging
import pytest
from unittest.mock import Mock
from horizon_data_core.logging import setup_extra_logger, log_calls


def test_setup_extra_logger() -> None:
    """Test that setup_extra_logger returns a logger."""
    logger = setup_extra_logger()
    assert isinstance(logger, logging.Logger)


def test_log_calls_exception() -> None:
    """Test that log_calls decorator logs exceptions and re-raises them."""
    mock_logger = Mock(spec=logging.Logger)
    
    @log_calls(mock_logger, logging.INFO)
    def failing_function() -> None:
        raise ValueError("Test error")
    
    # Call the decorated function and verify exception is raised
    with pytest.raises(ValueError, match="Test error"):
        failing_function()
    
    # Verify exception was logged
    assert mock_logger.log.called
    # Check that ERROR log was called with exception details
    error_calls = [call for call in mock_logger.log.call_args_list if "ERROR" in str(call)]
    assert len(error_calls) > 0


"""ORM types for the Horizon Data Core SDK."""

from collections.abc import Callable
from datetime import datetime as dt
from typing import Any

from sqlalchemy import Column, DateTime, Float, Integer, MetaData, String, Text, Uuid
from sqlalchemy.dialects.postgresql import ARRAY, CITEXT
from sqlalchemy.orm import Mapped, declarative_base, mapped_column
from sqlalchemy.types import Interval, UserDefinedType

schema_metadata = MetaData(schema="horizon_public")
HorizonPublicOrmBase = declarative_base(metadata=schema_metadata)


class HorizonPublicOrm(HorizonPublicOrmBase):
    """Abstract Base ORM Class for Horizon."""

    __abstract__ = True
    __table__ = None

    created_datetime: Mapped[dt] = mapped_column(DateTime, nullable=False)  # DB default is now()
    modified_datetime: Mapped[dt] = mapped_column(DateTime, nullable=False)  # DB default is now(), updates onupdate


class PostgreSQLPoint(UserDefinedType):
    """Custom type for PostgreSQL point.

    This serializes (longitude, latitude) pairs into "(lon,lat)" as required by the postgresql "POINT" type.

    The typing for this is incomplete, but that is due to how sqlAlchemy has created the
    UserDefinedType class from which this inherits.

    This type signature comes from SQLAlchemy's UserDefinedType class so we have to just
    ignore the Any type signature and the lint errors it throws for now.
    """

    cache_ok = True

    def get_col_spec(self, **kwargs: Any) -> str:  # noqa: ANN401, ARG002
        """Get the column specification for the point."""
        return "POINT"

    def bind_processor(self, dialect: Any) -> Callable[[tuple[float, float] | None | Any], str | None | Any]:  # noqa: ANN401, ARG002 # pyright: ignore[reportIncompatibleMethodOverride]
        """Bind the value to the database."""

        def process(value: tuple[float, float] | None | Any) -> str | None | Any:  # noqa: ANN401
            if value is None:
                return None
            if isinstance(value, tuple):
                return f"({value[0]},{value[1]})"
            return value

        return process

    def result_processor(self, dialect: Any, coltype: Any) -> Callable[[Any], Any]:  # noqa: ANN401, ARG002 # pyright: ignore[reportIncompatibleMethodOverride]
        """Process the value from the database."""

        def process(value: Any) -> Any:  # noqa: ANN401
            if value is None:
                return None
            if isinstance(value, str):
                # Parse "(x,y)" format
                coords = value.strip("()").split(",")
                return (float(coords[0]), float(coords[1]))
            return value

        return process


class EntityOrm(HorizonPublicOrm):
    """An instance of an entity, used for SQL ORM."""

    __tablename__ = "entity"

    id = Column(Uuid, primary_key=True)
    name = Column(CITEXT)
    kind_id = Column(Uuid)
    free_text = Column(Text)
    position = Column(PostgreSQLPoint)
    start_datetime = Column(DateTime)
    end_datetime = Column(DateTime)
    organization_id = Column(Uuid)

    def __repr__(self) -> str:
        """Return a string representation of the EntityOrm."""
        return (
            f"EntityOrm(id={self.id}, name={self.name}, position={self.position}, "
            f"organization_id={self.organization_id}, created_datetime={self.created_datetime}, "
            f"modified_datetime={self.modified_datetime}, kind_id={self.kind_id}, free_text={self.free_text})"
        )


class DataStreamOrm(HorizonPublicOrm):
    """A data stream of an entity."""

    __tablename__ = "data_stream"

    id = Column(Uuid, primary_key=True)
    entity_id = Column(Uuid)
    organization_id = Column(Uuid)
    name = Column(Text)


class DataRowOrm(HorizonPublicOrm):
    """A row of data in the data_row table.

    This is a timescaledb hypertable and requires slightly special handling.

    There is no id field or primary key as the data is segmented by partition and time.
    We can exclude the "id" field that is inherited from the base class.
    """

    __tablename__ = "data_row"

    data_stream_id = Column(Uuid, primary_key=True)
    datetime = Column(DateTime, primary_key=True)
    created_datetime: Mapped[dt] = mapped_column(DateTime, primary_key=True, default=dt.now)
    vector = Column(ARRAY(Float))
    data_type = Column(String, primary_key=True)
    track_id = Column(Uuid, primary_key=True)
    vector_start_bound = Column(Float)
    vector_end_bound = Column(Float)


class MetadataRowOrm(HorizonPublicOrm):
    """A row of data in the metadata_row table corresponding.

    This is a timescaledb hypertable and requires slightly special handling.

    There is no id field or primary key as the data is segmented by partition and time.
    We can exclude the "id" field that is inherited from the base class.
    """

    __tablename__ = "metadata_row"

    data_stream_id = Column(Uuid, primary_key=True)
    datetime = Column(DateTime, primary_key=True)
    created_datetime: Mapped[dt] = mapped_column(DateTime, primary_key=True, default=dt.now)
    latitude = Column(Float)
    longitude = Column(Float)
    altitude = Column(Float)
    speed = Column(Float)
    heading = Column(Float)
    pitch = Column(Float)
    roll = Column(Float)
    speed_over_ground = Column(Float)


class MissionOrm(HorizonPublicOrm):
    """A group that can group entities around some bounded time or purpose."""

    __tablename__ = "mission"

    id = Column(Uuid, primary_key=True)
    name = Column(CITEXT)
    start_datetime = Column(DateTime)
    end_datetime = Column(DateTime)
    position = Column(PostgreSQLPoint)
    free_text = Column(Text)
    organization_id = Column(Uuid)


class MissionEntityOrm(HorizonPublicOrm):
    """A relationship table that specifies entity association with a mission."""

    __tablename__ = "mission_entity"

    id = Column(Uuid, primary_key=True)
    mission_id = Column(Uuid)
    entity_id = Column(Uuid)
    organization_id = Column(Uuid)


class OntologyOrm(HorizonPublicOrm):
    """An ontology for organizing entities."""

    __tablename__ = "ontology"

    id = Column(Uuid, primary_key=True)
    name = Column(CITEXT)
    description = Column(Text)
    organization_id = Column(Uuid)


class OntologyClassOrm(HorizonPublicOrm):
    """A class within an ontology."""

    __tablename__ = "ontology_class"

    id = Column(Uuid, primary_key=True)
    ontology_id = Column(Uuid)
    parent_id = Column(Uuid)
    name = Column(CITEXT)
    description = Column(Text)
    order = Column(Integer)
    organization_id = Column(Uuid)
    relationship_type = Column(CITEXT)


class BeamgramSpecificationOrm(HorizonPublicOrm):
    """A specification for beamgram data processing."""

    __tablename__ = "beamgram_specification"

    id = Column(Uuid, primary_key=True)
    center_bearing = Column(Float)
    center_bin_width = Column(Float)
    min_frequency = Column(Float)
    max_frequency = Column(Float)
    nfft = Column(Integer)
    organization_id = Column(Uuid)
    elevation = Column(Float)
    update_rate = Column(Interval)
    normalizer = Column(Text)


class BearingTimeRecordSpecificationOrm(HorizonPublicOrm):
    """A specification for bearing-time record data processing."""

    __tablename__ = "bearing_time_record_specification"

    id = Column(Uuid, primary_key=True)
    min_frequency = Column(Float)
    max_frequency = Column(Float)
    nfft = Column(Integer)
    frequency_spacing = Column(Float)
    organization_id = Column(Uuid)
    elevation = Column(Float)
    update_rate = Column(Interval)
    normalizer = Column(Text)


class EntityBeamgramSpecificationOrm(HorizonPublicOrm):
    """A relationship table that specifies entity association with a beamgram specification."""

    __tablename__ = "entity_beamgram_specification"

    id = Column(Uuid, primary_key=True)
    entity_id = Column(Uuid)
    beamgram_specification_id = Column(Uuid)
    organization_id = Column(Uuid)


class EntityBearingTimeRecordSpecificationOrm(HorizonPublicOrm):
    """A relationship table that specifies entity association with a btr specification."""

    __tablename__ = "entity_bearing_time_record_specification"

    id = Column(Uuid, primary_key=True)
    entity_id = Column(Uuid)
    bearing_time_record_specification_id = Column(Uuid)
    organization_id = Column(Uuid)



"""SDK for Horizon Data Core operations."""
# Do not lint for exception strings being assigned to variables first
# ruff: noqa: EM101
# ruff: noqa: EM102
# Do not lint for logger f-strings
# ruff: noqa: G004
# Do not lint for long messages in exceptions
# ruff: noqa: TRY003

import logging
import random
import time
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import UTC, datetime, timedelta
from typing import Any
from uuid import UUID

import pyarrow as pa
from pydantic import BaseModel
from pyiceberg.catalog import Catalog, load_catalog
from pyiceberg.exceptions import CommitFailedException, OAuthError
from pyiceberg.table import Table as IcebergTable
from sqlalchemy import inspect, select
from sqlalchemy.dialects.postgresql import insert

from horizon_data_core.logging import log_calls

from .base_types import (
    BasePostgresModel,
    BasePyarrowModel,
    BeamgramSpecification,
    BearingTimeRecordSpecification,
    DataRow,
    DataStream,
    Entity,
    EntityBeamgramSpecification,
    EntityBearingTimeRecordSpecification,
    MetadataRow,
    Mission,
    MissionEntity,
    Ontology,
    OntologyClass,
)
from .client import Client
from .orm_types import (
    BeamgramSpecificationOrm,
    BearingTimeRecordSpecificationOrm,
    DataRowOrm,
    DataStreamOrm,
    EntityBeamgramSpecificationOrm,
    EntityBearingTimeRecordSpecificationOrm,
    EntityOrm,
    HorizonPublicOrm,
    MetadataRowOrm,
    MissionEntityOrm,
    MissionOrm,
    OntologyClassOrm,
    OntologyOrm,
)

logger = logging.getLogger(__name__)


class BaseRepository[B: BaseModel](ABC):
    """Base repository for database operations.

    Defines a basic set of operations:
        - insert        Create a new database entry
        - insert_batch  Create several new database entries
        - upsert        Create (or update) a new database entry (if matching entry exists)
        - upsert_batch  Create (or update) several new entries (if matching entry exists)
        - read          Return the entry given its id
        - update        Update an existing entry, error if it does not already exist
        - delete        Remove an existing entry
        - list          Return a list of entries (possibly empty) from a list of filter conditions
    """

    def __init__(self, client: Client) -> None:
        """Initialize the base repository.

        Args:
            client: Database client for operations
        """
        self.client = client

    @abstractmethod
    def insert(self, model: B) -> B:
        """Insert a new record."""

    @abstractmethod
    def insert_batch(self, models: list[B]) -> list[B]:
        """Create a new record batch."""

    @abstractmethod
    def upsert(self, model: B, unique_fields: list[str], merge_fields: list[str]) -> B:
        """Insert a new record or update if it already exists."""

    @abstractmethod
    def upsert_batch(self, models: list[B], unique_fields: list[str], merge_fields: list[str]) -> list[B]:
        """Insert a new record or update if it already exists."""

    @abstractmethod
    def read(self, id: UUID) -> B | None:
        """Read a record by ID."""

    @abstractmethod
    def update(self, model: B) -> B:
        """Update an existing record."""

    @abstractmethod
    def delete(self, id: UUID) -> bool:
        """Delete a record by ID."""

    @abstractmethod
    def list(self, **filters: Any) -> list[B]:  # noqa: ANN401
        """List records with optional filters."""


class PostgresRepository[P: BasePostgresModel, O: HorizonPublicOrm](BaseRepository[P]):
    """Repository for PostgreSQL operations."""

    def __init__(
        self, model_class: type[P], orm_class: type[O], client: Client, organization_id: UUID | None = None
    ) -> None:
        """Initialize the PostgreSQL repository.

        Args:
            client: PostgreSQL client for database operations
            model_class: The model class this repository handles
            organization_id: Organization ID for multi-tenancy (optional)

        Raises:
            ValueError: If no ORM class is found for the model class
        """
        super().__init__(client)
        self.model_class = model_class
        self.organization_id = organization_id
        self.orm_class = orm_class

    def _base_model_to_orm(self, model: P) -> O:
        """Convert BaseModel to SQLAlchemy ORM model."""
        model_dict = model.model_dump(exclude_none=True)
        # Ensure organization_id is set if its part of the model
        if "organization_id" in self.model_class.model_fields:
            model_dict["organization_id"] = self.organization_id
        # Modified datetime is set by trigger on the database side, but has a NOT NULL constraint
        # (abuck) I couldn't figure out how to get around the NOT NULL constraint so we just set it to "now"
        # in the SDK and allow a user to send "None".
        model.modified_datetime = datetime.now(UTC)
        orm = self.orm_class(**model_dict)
        orm.modified_datetime = datetime.now(UTC)
        return orm

    def _orm_to_base_model(self, orm_instance: O) -> P:
        """Convert SQLAlchemy ORM model to BaseModel."""
        mapper = inspect(orm_instance.__class__)
        assert mapper is not None  # inspect will raise an exception rather than returning None
        orm_dict = {column.name: getattr(orm_instance, column.name) for column in mapper.columns}
        return self.model_class(**orm_dict)

    @log_calls(logger, level=logging.DEBUG)
    def insert(self, model: P) -> P:
        """Insert a new record in PostgreSQL.

        Args:
            model: The model instance to create

        Returns:
            The inserted model instance with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        with self.client.session() as session:
            # Convert all models to ORM instances
            orm_instance = self._base_model_to_orm(model)
            session.add(orm_instance)
            session.commit()
            # Convert back to models
            return self._orm_to_base_model(orm_instance)

    def insert_batch(self, models: list[P]) -> list[P]:
        """Create a new record batch in PostgreSQL.

        Args:
            models: List of model instances to create

        Returns:
            List of created model instances with updated fields or empty list if input is empty

        Raises:
            SQLAlchemyError: If database operation fails
        """
        # Early return for empty input
        if not models:
            logger.warning("Empty input provided to upsert_batch. No action taken.")
            return []

        with self.client.session() as session:
            # Convert all models to ORM instances
            orm_instances = [self._base_model_to_orm(m) for m in models]
            session.add_all(orm_instances)
            session.commit()
            # Convert back to models
            return [self._orm_to_base_model(orm) for orm in orm_instances]

    def upsert(
        self,
        model: P,
        unique_fields: list[str] | None = None,  # Natural key columns
        merge_fields: list[str] | None = None,  # Columns to update on conflict
    ) -> P:
        """
        Upsert using PostgreSQL's ON CONFLICT.

        Args:
            model: The model to upsert
            unique_fields: Columns that define uniqueness (e.g., ['name', 'mission_id'])
            merge_fields: Columns to update on conflict (None = all except conflict_columns)

        Returns:
            The created or updated model instance

        Raises:
            SQLAlchemyError: If database operation fails
        """
        with self.client.session() as session:
            orm_instance = self._base_model_to_orm(model)
            orm_class = self.orm_class

            # Get all columns
            inspector = inspect(orm_class)
            assert inspector is not None  # inspect will raise an exception rather than returning None
            table_columns = inspector.columns
            all_columns = table_columns.keys()

            # Default: use primary key as conflict column
            if unique_fields is None:
                unique_fields = [c.key for c in table_columns if c.primary_key]

            # Default: update all columns except conflict columns
            if merge_fields is None:
                merge_fields = [c for c in all_columns if c not in unique_fields]

            # Build values dict
            values = {c: getattr(orm_instance, c, None) for c in all_columns}

            # Build statement
            insert_stmt = insert(orm_class).values(**values)
            stmt = insert_stmt.on_conflict_do_update(
                index_elements=unique_fields, set_={col: getattr(insert_stmt.excluded, col) for col in merge_fields}
            ).returning(orm_class)

            result = session.execute(stmt)
            merged_instance = result.scalar_one()

            session.commit()
            session.refresh(merged_instance)
            return self._orm_to_base_model(merged_instance)

    def upsert_batch(
        self,
        models: list[P],
        unique_fields: list[str] | None = None,
        merge_fields: list[str] | None = None,
    ) -> list[P]:
        """
        Batch upsert using PostgreSQL's ON CONFLICT.

        Args:
            models: List of models to upsert
            unique_fields: Columns that define uniqueness
            merge_fields: Columns to update on conflict
        """
        # Early return for empty input
        if not models:
            logger.warning("Empty input provided to upsert_batch. No action taken.")
            return []

        with self.client.session() as session:
            # Convert all models to ORM instances
            orm_instances = [self._base_model_to_orm(m) for m in models]
            orm_class = self.orm_class

            # Get all columns
            inspector = inspect(orm_class)
            assert inspector is not None  # inspect will raise an exception rather than returning None
            table_columns = inspector.columns
            all_columns = table_columns.keys()

            # Default: use primary key as conflict column
            if unique_fields is None:
                unique_fields = [c.key for c in table_columns if c.primary_key]

            if merge_fields is None:
                merge_fields = [c for c in all_columns if c not in unique_fields]

            # Build list of value dicts
            values_list = [{col: getattr(orm, col, None) for col in all_columns} for orm in orm_instances]

            # Build INSERT statement with multiple rows
            insert_stmt = insert(orm_class).values(values_list)
            stmt = insert_stmt.on_conflict_do_update(
                index_elements=unique_fields, set_={col: getattr(insert_stmt.excluded, col) for col in merge_fields}
            ).returning(orm_class)

            # Execute and get all results
            result = session.execute(stmt)
            merged_instances = result.scalars().all()

            session.commit()

            # Convert back to models
            return [self._orm_to_base_model(orm) for orm in merged_instances]

    def read(self, id: UUID) -> P | None:
        """Read a record by ID from PostgreSQL.

        Args:
            id: The UUID of the record to read

        Returns:
            The model instance if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        with self.client.session() as session:
            orm_instance = session.get(self.orm_class, id)
            if orm_instance:
                return self._orm_to_base_model(orm_instance)
            return None

    def update(self, model: P) -> P:
        """Update an existing record in PostgreSQL.

        Args:
            model: The model instance with updated fields

        Returns:
            The updated model instance

        Raises:
            ValueError: If record with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        with self.client.session() as session:
            orm_instance = session.get(self.orm_class, model.id)
            if not orm_instance:
                raise ValueError(f"Record with ID {model.id} not found")

            # Update fields
            model_dict = model.model_dump(exclude_none=True)
            for key, value in model_dict.items():
                if hasattr(orm_instance, key):
                    setattr(orm_instance, key, value)

            session.commit()
            session.refresh(orm_instance)
            return self._orm_to_base_model(orm_instance)

    def delete(self, id: UUID) -> bool:
        """Delete a record by ID from PostgreSQL.

        Args:
            id: The UUID of the record to delete

        Returns:
            True if the record was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        with self.client.session() as session:
            orm_instance = session.get(self.orm_class, id)
            if orm_instance:
                session.delete(orm_instance)
                session.commit()
                return True
            return False

    def list(self, **filters: dict[str, Any]) -> list[P]:
        """List records with optional filters from PostgreSQL.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of model instances matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        with self.client.session() as session:
            query = select(self.orm_class)

            # Apply filters
            for key, value in filters.items():
                if hasattr(self.orm_class, key):
                    query = query.where(getattr(self.orm_class, key) == value)

            orm_instances = session.execute(query).scalars().all()
            return [self._orm_to_base_model(instance) for instance in orm_instances]


class IcebergRepository[A: BasePyarrowModel](BaseRepository[A]):  # pragma: no cover
    """Repository for Iceberg table operations."""

    def __init__(
        self,
        catalog: Catalog,
        refresh_catalog: Callable[[], Catalog],
        table_name: str,
        model_class: type[A],
        organization_id: UUID | None = None,
        batch_size: int = 10,
        max_retries: int = 3,
    ) -> None:
        """Initialize the Iceberg repository.

        Args:
            catalog: Iceberg catalog for table operations
            table_name: Name of the Iceberg table
            model_class: The model class this repository handles
            organization_id: Organization ID for multi-tenancy (optional)
            batch_size: Batch size for buffered writes (default: 10)
            max_retries: Maximum retry attempts for failed operations (default: 3)
        """
        self.catalog = catalog
        self.table_name = table_name
        self.model_class = model_class
        self.organization_id = organization_id
        self.table_name = table_name
        self.table = self.catalog.load_table(self.table_name)
        self.buffer: dict[str, list[A]] = {}
        self.batch_size = batch_size
        self.max_retries = max_retries
        self.refresh_catalog = refresh_catalog
        self.catalog_ttl = timedelta(minutes=1)
        self.catalog_created_at = datetime.now(tz=UTC)

    def _append(self, pyarrow_table: pa.Table, max_retries: int | None = None) -> None:
        """Append a PyArrow table to the Iceberg table with retry logic.

        Uses exponential backoff with jitter for retry attempts.

        Args:
            pyarrow_table: The PyArrow table to append
            max_retries: Maximum number of retry attempts (uses instance default if None)

        Raises:
            CommitFailedException: If all retry attempts fail
        """
        if max_retries is None:
            max_retries = self.max_retries
        """Append a PyArrow table to the Iceberg table."""
        retry_count = 0
        sleep_time = 5.0  # 5 seconds
        exponential_backoff_factor = 2.0  # 2x the previous sleep time
        jitter_factor = (
            0.5  # jitter factor on backoff_factor [0 => no jitter, jitter over full range of backoff_factor]
        )
        assert jitter_factor > 0, "Jitter factor should be greater than 0 for algorithm stability."
        assert jitter_factor < exponential_backoff_factor, (
            "Jitter factor should be less than the exponential backoff factor for algorithm stability."
        )
        max_sleep_time = 1200.0  # max sleep time in seconds
        while retry_count < max_retries:
            try:
                logger.info("Appending PyArrow table to Iceberg table, attempt %s of %s", retry_count + 1, max_retries)
                start_time = time.time()
                self.refresh_table().append(pyarrow_table)
            except CommitFailedException as e:
                # We are not yet re-raising the exception so we are not logging it as an exception yet...
                logger.error("Error appending PyArrow table to Iceberg table: %s", e)  # noqa: TRY400
                retry_count += 1
                logger.info("Sleeping for %s seconds before retrying", sleep_time)
                time.sleep(sleep_time)
                # Update backoff with jitter
                sleep_time = sleep_time * (
                    # random.random() is not being used for cryptographic purposes
                    # so it is suitable for this use case: S311
                    exponential_backoff_factor * (1.0 + random.random() * jitter_factor * 2.0 - jitter_factor)  # noqa: S311
                )
                sleep_time = min(sleep_time, max_sleep_time)
                if retry_count >= max_retries:
                    logger.exception("Failed to append pyarrow table to iceberg table after %s retries", max_retries)
                    raise
            else:
                # If we get here, we successfully appended the PyArrow table to the Iceberg table
                logger.info(
                    "Appended PyArrow table to Iceberg table, attempt %s of %s, time taken: %s seconds",
                    retry_count + 1,
                    max_retries,
                    time.time() - start_time,
                )
                return

    def refresh_table(self) -> IcebergTable:
        """Refresh the Iceberg table.

        Reloads the table from the catalog to get the latest metadata.

        Returns:
            The refreshed Iceberg table instance
        """
        try:
            self.table = self.catalog.load_table(self.table_name)
        except OAuthError as e:
            logger.warning("OAuthError when loading Iceberg table: %s", e)
            logger.warning("Attempting to refresh OAuth token by reloading the catalog")
            # Refresh the catalog
            self.catalog = self.refresh_catalog()
            # Reload the table
            self.table = self.catalog.load_table(self.table_name)
            logger.warning("Successfully refreshed Iceberg table")

        return self.table

    def _should_flush(self, buffer_name: str) -> bool:
        """Check if the buffer should be flushed.

        Args:
            buffer_name: Name of the buffer to check

        Returns:
            True if buffer size exceeds batch_size, False otherwise
        """
        buffer = self._get_buffer(buffer_name)
        return len(buffer) >= self.batch_size

    def _write(self, models: list[A]) -> None:
        """Write a list of models to the Iceberg table.

        Converts models to PyArrow format and appends to the table.

        Args:
            models: List of models to write

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        if not models:
            logger.debug("IcebergRepository._write called with empty models; skipping.")
            return

        table = pa.concat_tables([model.to_pyarrow() for model in models])
        self._append(table)

    def _clear_buffer(self, buffer_name: str) -> None:
        """Clear the buffer for a given buffer name.

        Args:
            buffer_name: Name of the buffer to clear
        """
        self.buffer[buffer_name] = []

    def _get_buffer(self, buffer_name: str) -> list[A]:
        """Get the buffer for a given buffer name.

        Creates a new empty buffer if it doesn't exist.

        Args:
            buffer_name: Name of the buffer to get

        Returns:
            List of model instances in the buffer
        """
        if buffer_name not in self.buffer:
            self.buffer[buffer_name] = []
        return self.buffer[buffer_name]

    def _pop_buffer(self, buffer_name: str) -> list[A]:
        """Pop the buffer for a given buffer name.

        Args:
            buffer_name: Name of the buffer to pop
        """
        if buffer_name not in self.buffer:
            return []
        return self.buffer.pop(buffer_name)

    def flush(self, buffer_name: str | None = None) -> None:
        """Flush the buffer to the Iceberg table.

        If no buffer_name is provided, all buffers will be flushed.

        Args:
            buffer_name: Name of specific buffer to flush, or None to flush all buffers

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        if buffer_name:
            logger.info(f"Flushing buffer: {buffer_name}")
            buffer = self._pop_buffer(buffer_name)
            self._write(buffer)
        else:
            self.group_flush(list(self.buffer.keys()))

    def group_flush(self, buffer_names: list[str]) -> None:
        """Flush a group of buffers to the Iceberg table.

        Args:
            buffer_names: List of buffer names to flush
        """
        if len(buffer_names) == 0:
            logger.debug("No buffers to flush")
            return
        group_buffer = [model for buffer_name in buffer_names for model in self._pop_buffer(buffer_name)]
        logger.info("Flushing buffers %s to Iceberg table", buffer_names)
        self._write(group_buffer)

    def tick(self) -> None:
        """Tick the Iceberg repository.

        This is used to flush the buffers if they are not flushed by the time the tick is called.
        """
        # Refresh the catalog if it has expired
        if datetime.now(tz=UTC) - self.catalog_created_at >= self.catalog_ttl:
            logger.info("Refreshing Iceberg catalog")
            self.catalog = self.refresh_catalog()
            self.catalog_created_at = datetime.now(tz=UTC)

        # Flush any buffers that are ready to be flushed as a group
        buffers_to_flush = []
        for buf_name in self.buffer:
            if self._should_flush(buf_name):
                logger.debug("(%s) Adding buffer to flush list", buf_name)
                buffers_to_flush.append(buf_name)
        self.group_flush(buffers_to_flush)

    def insert(self, model: A, buffer_name: str | None = None) -> A:
        """Create a new record in Iceberg table.

        Args:
            model: The model instance to create
            buffer_name: Optional buffer name for batching writes

        Returns:
            The model instance (unchanged)

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        if buffer_name:
            buffer = self._get_buffer(buffer_name)
            buffer.append(model)
            logger.debug(f"({buffer_name}) Buffer size now %s rows", len(buffer))
        else:
            self._write([model])
        return model

    def insert_batch(self, models: list[A], buffer_name: str | None = None) -> list[A]:
        """Create a new record batch in Iceberg table.

        If a buffer_name is provided, the models will be added to the buffer and flushed as needed.

        If no buffer_name is provided, the models will be written directly to the Iceberg table.
        NOTE: THIS WILL BYPASS THE BUFFER, so it may result in out-of-order writes if you mix
        buffered and non-buffered writes.

        Args:
            models: List of model instances to create
            buffer_name: Optional buffer name for batching writes

        Returns:
            The list of model instances (unchanged)

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        # Write to Iceberg table
        if buffer_name:
            buffer = self._get_buffer(buffer_name)
            buffer.extend(models)
            logger.debug(f"Buffer size now: {len(buffer)}")
        else:
            self._write(models)
        return models

    def upsert(self, model: A, unique_fields: list[str], merge_fields: list[str]) -> A:
        """Upsert a new record in Iceberg table.

        Args:
            model: The model instance to create or update
            unique_fields: The list of fields to detect conflict
            merge_fields: The fields to update on conflict

        Raises:
            NotImplementedError: Always raised as Iceberg doesn't support upserts
        """
        # Iceberg doesn't have native upsert support, so we just create
        raise NotImplementedError("Upsert operations for Iceberg tables not implemented")

    def read(self, id: UUID) -> A | None:
        """Read a record by ID from Iceberg table.

        Args:
            id: The UUID of the record to read

        Raises:
            NotImplementedError: Always raised as Iceberg doesn't support direct ID lookups
        """
        # Note: Iceberg doesn't have direct ID-based lookups like SQL
        raise NotImplementedError("Read operations for Iceberg tables not implemented")

    def update(self, model: A) -> A:
        """Update an existing record in Iceberg table.

        Args:
            model: The model instance with updated fields

        Raises:
            NotImplementedError: Always raised as Iceberg updates are table-specific
        """
        raise NotImplementedError("Update operations for Iceberg tables not implemented")

    def delete(self, id: UUID) -> bool:
        """Delete a record by ID from Iceberg table.

        Args:
            id: The UUID of the record to delete

        Raises:
            NotImplementedError: Always raised as Iceberg deletes are table-specific
        """
        raise NotImplementedError("Delete operations for Iceberg tables not implemented")

    def list(self, **filters: Any) -> list[A]:  # noqa: ANN401
        """List records with optional filters from Iceberg table.

        Args:
            **filters: Iceberg scan parameters including:
                - row_filter: str | boolean_expr = True
                - selected_fields: tuple[str] = ("*",)
                - case_sensitive: bool = True
                - snapshot_id: str | None = None
                - options: Properties = {}
                - limit: int | None = None

        Returns:
            List of model instances matching the filters

        Raises:
            Exception: If Iceberg scan operation fails
        """
        # Convert filters to Iceberg query format
        table = self.refresh_table().scan(**filters).to_polars()
        # A given instance of an Iceberg Repository is tied to a single table, which means
        # it is tied to a single schema and model_class, so we should be able to
        # read from the table and directly instantiate the model_class with the row
        return [self.model_class(**row) for row in table.iter_rows(named=True)]


class HorizonSDK:
    """Main SDK class for Horizon Data Core operations."""

    def __init__(
        self,
        postgres_client: Client,
        # Iceberg repositories were deprecated after version 0.3.6. This will be removed in version 0.5.0
        catalog_properties: dict[str, str | None],  # noqa: ARG002
        organization_id: UUID | None = None,
        # Iceberg repositories were deprecated after version 0.3.6. This will be removed in version 0.5.0
        iceberg_batch_size: int = 16,  # noqa: ARG002
        # Iceberg repositories were deprecated after version 0.3.6. This will be removed in version 0.5.0
        max_retries: int = 10,  # noqa: ARG002
    ) -> None:
        """Initialize the Horizon SDK.

        Args:
            postgres_client: PostgreSQL client for database operations
            catalog_properties: Properties for the Iceberg catalog
            organization_id: Organization ID for multi-tenancy (optional)
            iceberg_batch_size: (DEPRECATED) Batch size for Iceberg operations (default: 16)
            max_retries: (DEPRECATED) Maximum retry attempts for failed operations (default: 10)
        """
        self.catalog_properties: dict[str, str | None] = {}
        self.postgres_client = postgres_client
        self.organization_id = organization_id

        # Initialize repositories for PostgreSQL models
        self.entities = PostgresRepository(Entity, EntityOrm, postgres_client, organization_id)
        self.data_streams = PostgresRepository(DataStream, DataStreamOrm, postgres_client, organization_id)
        self.missions = PostgresRepository(Mission, MissionOrm, postgres_client, organization_id)
        self.mission_entities = PostgresRepository(MissionEntity, MissionEntityOrm, postgres_client, organization_id)
        self.ontologies = PostgresRepository(Ontology, OntologyOrm, postgres_client, organization_id)
        self.ontology_classes = PostgresRepository(OntologyClass, OntologyClassOrm, postgres_client, organization_id)
        self.beamgram_specifications = PostgresRepository(
            BeamgramSpecification, BeamgramSpecificationOrm, postgres_client, organization_id
        )
        self.entity_beamgram_specifications = PostgresRepository(
            EntityBeamgramSpecification, EntityBeamgramSpecificationOrm, postgres_client, organization_id
        )
        self.bearing_time_record_specifications: PostgresRepository[
            BearingTimeRecordSpecification, BearingTimeRecordSpecificationOrm
        ] = PostgresRepository(
            BearingTimeRecordSpecification, BearingTimeRecordSpecificationOrm, postgres_client, organization_id
        )
        self.entity_bearing_time_record_specifications = PostgresRepository(
            EntityBearingTimeRecordSpecification,
            EntityBearingTimeRecordSpecificationOrm,
            postgres_client,
            organization_id,
        )
        self.data_rows = PostgresRepository(DataRow, DataRowOrm, postgres_client, organization_id)
        self.metadata_rows = PostgresRepository(MetadataRow, MetadataRowOrm, postgres_client, organization_id)

    def refresh_iceberg_catalog(self) -> Catalog:
        """Refresh the Iceberg catalog.

        DEPRECATED after v0.3.6
        """
        self.iceberg_catalog = load_catalog("rest", **self.catalog_properties)
        return self.iceberg_catalog

    def tick(self) -> None:
        """Tick the Iceberg repository.

        DEPRECATED after version 0.3.6.
        """
        return

    # Entity-BeamgramSpecification operations
    def create_entity_beamgram_specification(
        self, entity_beamgram_specification: EntityBeamgramSpecification
    ) -> EntityBeamgramSpecification:
        """Create a new entity-beamgram specification relationship."""
        return self.entity_beamgram_specifications.insert(entity_beamgram_specification)

    def list_entity_beamgram_specifications(self, **filters: Any) -> list[EntityBeamgramSpecification]:  # noqa: ANN401
        """List entity-beamgram specification relationships with optional filters."""
        return self.entity_beamgram_specifications.list(**filters)

    # Entity-BearingTimeRecordSpecification operations
    def create_entity_bearing_time_record_specification(
        self, entity_bearing_time_record_specification: EntityBearingTimeRecordSpecification
    ) -> EntityBearingTimeRecordSpecification:
        """Create a new entity-btr specification relationship."""
        return self.entity_bearing_time_record_specifications.insert(entity_bearing_time_record_specification)

    def list_entity_bearing_time_record_specifications(
        self,
        **filters: Any,  # noqa: ANN401
    ) -> list[EntityBearingTimeRecordSpecification]:
        """List entity-btr specification relationships with optional filters."""
        return self.entity_bearing_time_record_specifications.list(**filters)

    # PostgreSQL operations
    def create_entity(self, entity: Entity) -> Entity:
        """Create a new entity.

        Args:
            entity: The entity instance to create

        Returns:
            The created entity with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        logger.debug(
            "SDK: Creating entity",
            extra={"entity": entity},
        )
        return self.entities.upsert(entity)

    def create_or_update_entity(self, entity: Entity) -> Entity:
        """Create a new entity or update if it already exists.

        Uses custom datetime logic for start_datetime and end_datetime fields:
        - Updates start_datetime if new value is earlier than existing
        - Updates end_datetime if new value is later than existing

        Args:
            entity: The entity instance to create or update

        Returns:
            The created or updated entity

        Raises:
            SQLAlchemyError: If database operation fails
        """
        logger.debug(
            "SDK: Creating or updating entity",
            extra={"entity": entity},
        )

        # Custom logic for start_datetime and end_datetime fields:
        # - Updates start_datetime if new value is earlier than existing
        # - Keep start_datetime if new value is None and existing value is not None
        # - Updates end_datetime if new value is later than existing
        # - Keep end_datetime if new value is None and existing value is not None
        if entity.id is not None:
            existing_entity = self.entities.read(entity.id)
            if existing_entity is not None:
                updated_entity = entity

                # Keep start_datetime if both values are set and new value is later than existing
                # Keep start_datetime if new value is None and existing value is not None
                if (
                    updated_entity.start_datetime is not None
                    and existing_entity.start_datetime is not None
                    and updated_entity.start_datetime > existing_entity.start_datetime
                ) or (updated_entity.start_datetime is None and existing_entity.start_datetime is not None):
                    updated_entity.start_datetime = existing_entity.start_datetime

                # Keep end_datetime if both values are set and new value is later than existing
                # Keep end_datetime if new value is None and existing value is not None
                if (
                    updated_entity.end_datetime is not None
                    and existing_entity.end_datetime is not None
                    and updated_entity.end_datetime < existing_entity.end_datetime
                ) or (updated_entity.end_datetime is None and existing_entity.end_datetime is not None):
                    updated_entity.end_datetime = existing_entity.end_datetime

                return self.entities.upsert(updated_entity)
        # If no existing entity or no ID, do regular upsert
        return self.entities.upsert(entity)

    def read_entity(self, id: UUID) -> Entity | None:
        """Read an entity by ID.

        Args:
            id: The UUID of the entity to read

        Returns:
            The entity if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.entities.read(id)

    def update_entity(self, entity: Entity) -> Entity:
        """Update an existing entity.

        Args:
            entity: The entity instance with updated fields

        Returns:
            The updated entity

        Raises:
            ValueError: If entity with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.entities.update(entity)

    def delete_entity(self, id: UUID) -> bool:
        """Delete an entity by ID.

        Args:
            id: The UUID of the entity to delete

        Returns:
            True if the entity was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.entities.delete(id)

    def list_entities(self, **filters: Any) -> list[Entity]:  # noqa: ANN401
        """List entities with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of entities matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.entities.list(**filters)

    def create_data_stream(self, data_stream: DataStream) -> DataStream:
        """Create a new data stream.

        Args:
            data_stream: The data stream instance to create

        Returns:
            The created data stream with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.data_streams.upsert(data_stream)

    def read_data_stream(self, id: UUID) -> DataStream | None:
        """Read a data stream by ID.

        Args:
            id: The UUID of the data stream to read

        Returns:
            The data stream if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.data_streams.read(id)

    def update_data_stream(self, data_stream: DataStream) -> DataStream:
        """Update an existing data stream.

        Args:
            data_stream: The data stream instance with updated fields

        Returns:
            The updated data stream

        Raises:
            ValueError: If data stream with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.data_streams.upsert(data_stream)

    def delete_data_stream(self, id: UUID) -> bool:
        """Delete a data stream by ID.

        Args:
            id: The UUID of the data stream to delete

        Returns:
            True if the data stream was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.data_streams.delete(id)

    def list_data_streams(self, **filters: Any) -> list[DataStream]:  # noqa: ANN401
        """List data streams with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of data streams matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.data_streams.list(**filters)

    def create_mission(self, mission: Mission) -> Mission:
        """Create a new mission.

        Args:
            mission: The mission instance to create

        Returns:
            The created mission with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.missions.insert(mission)

    def read_mission(self, id: UUID) -> Mission | None:
        """Read a mission by ID.

        Args:
            id: The UUID of the mission to read

        Returns:
            The mission if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.missions.read(id)

    def update_mission(self, mission: Mission) -> Mission:
        """Update an existing mission.

        Args:
            mission: The mission instance with updated fields

        Returns:
            The updated mission

        Raises:
            ValueError: If mission with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.missions.update(mission)

    def delete_mission(self, id: UUID) -> bool:
        """Delete a mission by ID.

        Args:
            id: The UUID of the mission to delete

        Returns:
            True if the mission was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.missions.delete(id)

    def list_missions(self, **filters: Any) -> list[Mission]:  # noqa: ANN401
        """List missions with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of missions matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.missions.list(**filters)

    def create_mission_entity(self, mission_entity: MissionEntity) -> MissionEntity:
        """Create a new mission-entity relationship.

        Args:
            mission_entity: The mission-entity relationship instance to create

        Returns:
            The created mission-entity relationship with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.mission_entities.insert(mission_entity)

    def read_mission_entity(self, id: UUID) -> MissionEntity | None:
        """Read a mission-entity relationship by ID.

        Args:
            id: The UUID of the mission-entity relationship to read

        Returns:
            The mission-entity relationship if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.mission_entities.read(id)

    def update_mission_entity(self, mission_entity: MissionEntity) -> MissionEntity:
        """Update an existing mission-entity relationship.

        Args:
            mission_entity: The mission-entity relationship instance with updated fields

        Returns:
            The updated mission-entity relationship

        Raises:
            ValueError: If mission-entity relationship with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.mission_entities.update(mission_entity)

    def delete_mission_entity(self, id: UUID) -> bool:
        """Delete a mission-entity relationship by ID.

        Args:
            id: The UUID of the mission-entity relationship to delete

        Returns:
            True if the mission-entity relationship was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.mission_entities.delete(id)

    def list_mission_entities(self, **filters: Any) -> list[MissionEntity]:  # noqa: ANN401
        """List mission-entity relationships with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of mission-entity relationships matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.mission_entities.list(**filters)

    def create_ontology(self, ontology: Ontology) -> Ontology:
        """Create a new ontology.

        Args:
            ontology: The ontology instance to create

        Returns:
            The created ontology with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontologies.upsert(ontology)

    def read_ontology(self, id: UUID) -> Ontology | None:
        """Read an ontology by ID.

        Args:
            id: The UUID of the ontology to read

        Returns:
            The ontology if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontologies.read(id)

    def update_ontology(self, ontology: Ontology) -> Ontology:
        """Update an existing ontology.

        Args:
            ontology: The ontology instance with updated fields

        Returns:
            The updated ontology

        Raises:
            ValueError: If ontology with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.ontologies.update(ontology)

    def delete_ontology(self, id: UUID) -> bool:
        """Delete an ontology by ID.

        Args:
            id: The UUID of the ontology to delete

        Returns:
            True if the ontology was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontologies.delete(id)

    def list_ontologies(self, **filters: Any) -> list[Ontology]:  # noqa: ANN401
        """List ontologies with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of ontologies matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontologies.list(**filters)

    def create_ontology_class(self, ontology_class: OntologyClass) -> OntologyClass:
        """Create a new ontology class.

        Args:
            ontology_class: The ontology class instance to create

        Returns:
            The created ontology class with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontology_classes.upsert(ontology_class)

    def read_ontology_class(self, id: UUID) -> OntologyClass | None:
        """Read an ontology class by ID.

        Args:
            id: The UUID of the ontology class to read

        Returns:
            The ontology class if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontology_classes.read(id)

    def update_ontology_class(self, ontology_class: OntologyClass) -> OntologyClass:
        """Update an existing ontology class.

        Args:
            ontology_class: The ontology class instance with updated fields

        Returns:
            The updated ontology class

        Raises:
            ValueError: If ontology class with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.ontology_classes.update(ontology_class)

    def delete_ontology_class(self, id: UUID) -> bool:
        """Delete an ontology class by ID.

        Args:
            id: The UUID of the ontology class to delete

        Returns:
            True if the ontology class was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontology_classes.delete(id)

    def list_ontology_classes(self, **filters: Any) -> list[OntologyClass]:  # noqa: ANN401
        """List ontology classes with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of ontology classes matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.ontology_classes.list(**filters)

    # BeamgramSpecification operations
    def create_beamgram_specification(self, beamgram_specification: BeamgramSpecification) -> BeamgramSpecification:
        """Create a new beamgram specification.

        Args:
            beamgram_specification: The beamgram specification instance to create

        Returns:
            The created beamgram specification with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.beamgram_specifications.upsert(beamgram_specification)

    def create_or_update_beamgram_specification(
        self, beamgram_specification: BeamgramSpecification
    ) -> BeamgramSpecification:
        """Create a new beamgram specification or update if it already exists.

        Args:
            beamgram_specification: The beamgram specification instance to create or update

        Returns:
            The created or updated beamgram specification

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.beamgram_specifications.upsert(beamgram_specification)

    def read_beamgram_specification(self, id: UUID) -> BeamgramSpecification | None:
        """Read a beamgram specification by ID.

        Args:
            id: The UUID of the beamgram specification to read

        Returns:
            The beamgram specification if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.beamgram_specifications.read(id)

    def update_beamgram_specification(self, beamgram_specification: BeamgramSpecification) -> BeamgramSpecification:
        """Update an existing beamgram specification.

        Args:
            beamgram_specification: The beamgram specification instance with updated fields

        Returns:
            The updated beamgram specification

        Raises:
            ValueError: If beamgram specification with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.beamgram_specifications.update(beamgram_specification)

    def delete_beamgram_specification(self, id: UUID) -> bool:
        """Delete a beamgram specification by ID.

        Args:
            id: The UUID of the beamgram specification to delete

        Returns:
            True if the beamgram specification was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.beamgram_specifications.delete(id)

    def list_beamgram_specifications(self, **filters: Any) -> list[BeamgramSpecification]:  # noqa: ANN401
        """List beamgram specifications with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of beamgram specifications matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.beamgram_specifications.list(**filters)

    # BearingTimeRecordSpecification operations
    def create_bearing_time_record_specification(
        self, bearing_time_record_specification: BearingTimeRecordSpecification
    ) -> BearingTimeRecordSpecification:
        """Create a new bearing-time record specification.

        Args:
            bearing_time_record_specification: The bearing-time record specification instance to create

        Returns:
            The created bearing-time record specification with updated fields (e.g., generated ID)

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.bearing_time_record_specifications.upsert(bearing_time_record_specification)

    def create_or_update_bearing_time_record_specification(
        self, bearing_time_record_specification: BearingTimeRecordSpecification
    ) -> BearingTimeRecordSpecification:
        """Create a new bearing-time record specification or update if it already exists.

        Args:
            bearing_time_record_specification: The bearing-time record specification instance to create or update

        Returns:
            The created or updated bearing-time record specification

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.bearing_time_record_specifications.upsert(bearing_time_record_specification)

    def read_bearing_time_record_specification(self, id: UUID) -> BearingTimeRecordSpecification | None:
        """Read a bearing-time record specification by ID.

        Args:
            id: The UUID of the bearing-time record specification to read

        Returns:
            The bearing-time record specification if found, None otherwise

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.bearing_time_record_specifications.read(id)

    def update_bearing_time_record_specification(
        self, bearing_time_record_specification: BearingTimeRecordSpecification
    ) -> BearingTimeRecordSpecification:
        """Update an existing bearing-time record specification.

        Args:
            bearing_time_record_specification: The bearing-time record specification instance with updated fields

        Returns:
            The updated bearing-time record specification

        Raises:
            ValueError: If bearing-time record specification with the given ID is not found
            SQLAlchemyError: If database operation fails
        """
        return self.bearing_time_record_specifications.update(bearing_time_record_specification)

    def delete_bearing_time_record_specification(self, id: UUID) -> bool:
        """Delete a bearing-time record specification by ID.

        Args:
            id: The UUID of the bearing-time record specification to delete

        Returns:
            True if the bearing-time record specification was deleted, False if it didn't exist

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.bearing_time_record_specifications.delete(id)

    def list_bearing_time_record_specifications(self, **filters: Any) -> list[BearingTimeRecordSpecification]:  # noqa: ANN401
        """List bearing-time record specifications with optional filters.

        Args:
            **filters: Keyword arguments where keys are field names and values are filter values

        Returns:
            List of bearing-time record specifications matching the filters

        Raises:
            SQLAlchemyError: If database operation fails
        """
        return self.bearing_time_record_specifications.list(**filters)

    # Iceberg operations removed after v0.3.6. buffer_name parameter is deprecated and will be removed in v0.5.0
    def create_data_row(self, data_row: DataRow, buffer_name: str | None = None) -> DataRow:  # noqa: ARG002
        """Create a new data row in data_row table.

        Args:
            data_row: The data row instance to create
            buffer_name: (DEPRECATED) Optional buffer name for batching writes

        Returns:
            The data row instance (unchanged)

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        return self.data_rows.insert(data_row)

    def create_data_row_batch(self, data_rows: list[DataRow], buffer_name: str | None = None) -> list[DataRow]:  # noqa: ARG002
        """Create a new data row batch in data_row table.

        Args:
            data_rows: The list of data row instances to create
            buffer_name: (DEPRECATED) Optional buffer name for batching writes

        Returns:
            The data row instance (unchanged)

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        return self.data_rows.insert_batch(data_rows)

    def list_data_rows(self, **filters: Any) -> list[DataRow]:  # noqa: ANN401
        """List data rows with optional filters from Iceberg table.

        Args:
            **filters: Iceberg scan parameters including:

        Returns:
            List of data rows matching the filters

        Raises:
            Exception: If Iceberg scan operation fails
        """
        return self.data_rows.list(**filters)

    # Iceberg operations removed after v0.3.6. buffer_name parameter is deprecated and will be removed in v0.5.0
    def create_metadata_row(self, metadata_row: MetadataRow, buffer_name: str | None = None) -> MetadataRow:  # noqa: ARG002
        """Create a new metadata row in Iceberg table.

        Args:
            metadata_row: The metadata row instance to create
            buffer_name: (DEPRECATED) Optional buffer name for batching writes

        Returns:
            The metadata row instance (unchanged)

        Raises:
            CommitFailedException: If write operation fails after retries
        """
        return self.metadata_rows.insert(metadata_row)

    def list_metadata_rows(self, **filters: Any) -> list[MetadataRow]:  # noqa: ANN401
        """List metadata rows with optional filters from Iceberg table.

        Args:
            **filters: Postgres query values parameters including:

        Returns:
            List of metadata rows matching the filters

        Raises:
            Exception: If Iceberg scan operation fails
        """
        return self.metadata_rows.list(**filters)
